"""
widget specific commands
"""

import logging

from sen.tui.commands.base import register_command, SameThreadCommand


import urwidtrees


logger = logging.getLogger(__name__)


@register_command
class NavigateTopCommand(SameThreadCommand):
    name = "navigate-top"
    description = "go to first line"

    def run(self):
        # FIXME: refactor
        if isinstance(self.buffer.widget, urwidtrees.TreeBox):
            self.buffer.widget.focus_first()
        else:
            self.buffer.widget.set_focus(0)
            self.buffer.widget.reload_widget()


@register_command
class NavigateBottomCommand(SameThreadCommand):
    name = "navigate-bottom"
    description = "go to last line"

    def run(self):
        # FIXME: refactor
        if isinstance(self.buffer.widget, urwidtrees.TreeBox):
            self.buffer.widget.focus_last()
        else:
            self.buffer.widget.set_focus(len(self.buffer.widget.body) - 1)
            self.buffer.widget.reload_widget()


@register_command
class NavigateUpCommand(SameThreadCommand):
    name = "navigate-up"
    description = "go one line up"

    def run(self):
        return super(self.buffer.widget.__class__, self.buffer.widget).keypress(self.size, "up")


@register_command
class NavigateDownCommand(SameThreadCommand):
    name = "navigate-down"
    description = "go one line down"

    def run(self):
        return super(self.buffer.widget.__class__, self.buffer.widget).keypress(self.size, "down")


@register_command
class NavigateUpwardsCommand(SameThreadCommand):
    name = "navigate-upwards"
    description = "go 10 lines up"

    def run(self):
        if isinstance(self.buffer.widget, urwidtrees.TreeBox):
            self.ui.notify_message("This movement is not available.", level="error")
            return
        try:
            self.buffer.widget.set_focus(self.buffer.widget.get_focus()[1] - 10)
        except IndexError:
            self.buffer.widget.set_focus(0)
        self.buffer.widget.reload_widget()
        return


@register_command
class NavigateDownwardsCommand(SameThreadCommand):
    name = "navigate-downwards"
    description = "go 10 lines down"

    def run(self):
        if isinstance(self.buffer.widget, urwidtrees.TreeBox):
            self.ui.notify_message("This movement is not available.", level="error")
            return
        try:
            self.buffer.widget.set_focus(self.buffer.widget.get_focus()[1] + 10)
        except IndexError:
            self.buffer.widget.set_focus(len(self.buffer.widget.body) - 1)
        self.buffer.widget.reload_widget()
        return



import logging

import urwid


logger = logging.getLogger(__name__)


class ResponsiveColumns(urwid.Columns):
    """
    Widgets arranged horizontally in columns from left to right
    """

    def column_widths(self, size, focus=False):
        """
        Return a list of column widths.

        0 values in the list mean hide corresponding column completely
        """
        maxcol = size[0]
        self._cache_maxcol = maxcol
        widths = [width for i, (w, (t, width, b)) in enumerate(self.contents)]
        self._cache_column_widths = widths

        return widths



import urwid

from sen.tui.views.base import View
from sen.tui.widgets.list.util import SingleTextRow
from sen.tui.widgets.table import assemble_rows
from sen.tui.constants import MAIN_LIST_FOCUS
from sen.tui.widgets.util import SelectableText, get_map
from sen.tui.widgets.list.base import WidgetBase


class HelpBufferView(WidgetBase, View):
    def __init__(self, ui, buffer, global_keybinds):
        """

        :param ui:
        :param buffer: Buffer instance, display help about this buffer
        :param global_keybinds: dict
        """
        self.ui = ui
        self.buffer = buffer
        self.walker = urwid.SimpleFocusListWalker([])
        self.global_keybinds = global_keybinds
        super().__init__(ui, self.walker)

    def refresh(self):
        template = [
            SingleTextRow("Buffer: " + self.buffer.display_name,
                          maps={"normal": "main_list_white", "focus": MAIN_LIST_FOCUS}),
            SingleTextRow(""),
            SingleTextRow(self.buffer.description,
                          maps={"normal": "main_list_dg", "focus": MAIN_LIST_FOCUS}),
            SingleTextRow(""),
        ]

        if self.buffer.keybinds:
            template += [
                SingleTextRow("Buffer-specific Keybindings",
                              maps={"normal": "main_list_white", "focus": MAIN_LIST_FOCUS}),
                SingleTextRow(""),
            ]
            template.extend(assemble_rows(
                [[SelectableText(key, maps=get_map("main_list_yellow")),
                  SelectableText(command, maps=get_map("main_list_lg"))]
                 for key, command in self.buffer.keybinds.items()],
                ignore_columns=[1], dividechars=3))

        template += [
            SingleTextRow(""),
            SingleTextRow("Global Keybindings",
                          maps={"normal": "main_list_white", "focus": MAIN_LIST_FOCUS}),
            SingleTextRow("")
        ]
        template.extend(assemble_rows(
            [[SelectableText(key, maps=get_map("main_list_yellow")),
              SelectableText(command, maps=get_map("main_list_lg"))]
             for key, command in self.global_keybinds.items()],
            ignore_columns=[1], dividechars=3))

        self.set_body(template)
        self.set_focus(0)


class HelpCommandView(WidgetBase, View):
    def __init__(self, ui, command):
        """

        :param ui:
        :param command: Command instance
        """
        self.ui = ui
        self.command = command
        self.walker = urwid.SimpleFocusListWalker([])
        super().__init__(ui, self.walker)

    def refresh(self):
        invocation = ["[{}=value]".format(a.name) for a in self.command.options_definitions] + \
                     ["<{}>".format(o.name) for o in self.command.arguments_definitions]
        template = [
            SingleTextRow("Command: {} {}".format(self.command.name, " ".join(invocation)),
                          maps={"normal": "main_list_white", "focus": MAIN_LIST_FOCUS}),
            SingleTextRow(""),
        ]
        template += [SingleTextRow(s, maps={"normal": "main_list_dg", "focus": MAIN_LIST_FOCUS})
                     for s in self.command.description.split("\n")]
        template += [SingleTextRow("")]

        if self.command.arguments_definitions:
            template += [
                SingleTextRow(""),
                SingleTextRow("Arguments",
                              maps={"normal": "main_list_white", "focus": MAIN_LIST_FOCUS}),
                SingleTextRow(""),
            ]
            template.extend(assemble_rows(
                [[SelectableText(argument.name, maps=get_map("main_list_yellow")),
                  SelectableText(argument.description, maps=get_map("main_list_lg"))]
                 for argument in self.command.arguments_definitions],
                ignore_columns=[1], dividechars=3))

        if self.command.options_definitions:
            template += [
                SingleTextRow(""),
                SingleTextRow("Options",
                              maps={"normal": "main_list_white", "focus": MAIN_LIST_FOCUS}),
                SingleTextRow(""),
            ]
            template.extend(assemble_rows(
                [[SelectableText(option.name, maps=get_map("main_list_yellow")),
                  SelectableText(option.description, maps=get_map("main_list_lg"))]
                 for option in self.command.options_definitions],
                ignore_columns=[1], dividechars=3))

        self.set_body(template)
        self.set_focus(0)



import string
import random

import urwid


characters = string.digits + string.ascii_letters + string.punctuation


def get_random_text(length=32):
    return "".join([random.choice(characters) for _ in range(length)])


def get_random_text_widget(length=32):
    return urwid.Text(get_random_text(length=length))



from sen.docker_backend import DockerBackend, DockerContainer
from sen.tui.views.container_info import ProcessList
from tests.real import mock


def test_short_id():
    mock()
    b = DockerBackend()
    operation = DockerContainer({"Status": "Up", "Id": "voodoo"}, b).top()
    top_response = operation.response
    pt = ProcessList(top_response)

    # 24502
    #  \ 24542
    #     \ 23743
    #        \ 18725
    #        \ 18733
    #        \ 18743
    #        \ 23819

    root_process = pt.get_root_process()
    assert root_process.pid == "24502"
    assert pt.get_parent_process(root_process) is None

    p_24542 = pt.get_first_child_process(root_process)
    assert p_24542.pid == "24542"
    assert pt.get_last_child_process(root_process).pid == "24542"

    p_23743 = pt.get_first_child_process(p_24542)
    assert p_23743.pid == "23743"
    assert pt.get_last_child_process(p_24542).pid == "23743"

    p_18725 = pt.get_first_child_process(p_23743)
    assert p_18725.pid == "18725"
    assert pt.get_prev_sibling(p_18725) is None
    assert pt.get_parent_process(p_18725).pid == "23743"

    p_18733 = pt.get_next_sibling(p_18725)
    assert p_18733.pid == "18733"

    p_23819 = pt.get_last_child_process(p_23743)
    assert p_23819.pid == "23819"
    assert pt.get_next_sibling(p_23819) is None
    assert pt.get_parent_process(p_23819).pid == "23743"

    p_18743 = pt.get_prev_sibling(p_23819)
    assert p_18743.pid == "18743"

    assert pt.get_prev_sibling(p_18733) is p_18725
    assert pt.get_next_sibling(p_18733) is p_18743

    assert pt.get_prev_sibling(p_18743) is p_18733
    assert pt.get_next_sibling(p_18743) is p_23819



"""
Unsorted chunks.
"""

import logging

from sen.docker_backend import DockerImage, DockerContainer
from sen.tui.chunks.container import get_detailed_container_row
from sen.tui.chunks.image import get_detailed_image_row


logger = logging.getLogger(__name__)


def get_row(docker_object):
    if isinstance(docker_object, DockerImage):
        return get_detailed_image_row(docker_object)
    elif isinstance(docker_object, DockerContainer):
        return get_detailed_container_row(docker_object)
    else:
        raise Exception("what ")



import logging
import random
from itertools import chain

import pytest
from flexmock import flexmock
from urwid.listbox import SimpleListWalker

from sen.tui.widgets.list.base import WidgetBase
from sen.tui.widgets.list.common import ScrollableListBox, AsyncScrollableListBox
from sen.tui.widgets.list.util import ResponsiveRowWidget
from sen.tui.widgets.table import ResponsiveTable, assemble_rows
from .utils import get_random_text_widget
from .constants import SCREEN_WIDTH, SCREEN_HEIGHT


class MockUI:
    buffers = []

    def refresh(self):
        pass

    def set_alarm_in(self, *args, **kwargs):
        pass


class DataGenerator:
    @classmethod
    def text(cls, prefix="line", lines_no=3, return_bytes=False):
        s = "\n".join(["{}{}".format(prefix, x+1) for x in range(lines_no)])
        if return_bytes:
            return s.encode("utf-8")
        return s

    @classmethod
    def stream(cls, prefix="line", lines_no=3, return_bytes=False):
        text = []
        for x in range(lines_no):
            l = "{}{}\n".format(prefix, x+1)
            if return_bytes:
                l = l.encode("utf-8")
            text.append(l)
        s = chain(text)
        return iter(s)

    @classmethod
    def render(cls, prefix="line", lines_no=3, return_bytes=True):
        w = "{:%d}" % SCREEN_WIDTH
        response = []
        for x in range(SCREEN_HEIGHT):
            if x >= lines_no:
                l = w.format("")
            else:
                l = w.format("{}{}".format(prefix, x+1))
            if return_bytes:
                response.append(l.encode("utf-8"))
            else:
                response.append(l)
        return response


@pytest.mark.parametrize("inp,expected", [
    (DataGenerator.text(), DataGenerator.render()),
    (DataGenerator.text(return_bytes=True), DataGenerator.render()),
    (DataGenerator.text(prefix="liné"), DataGenerator.render(prefix="liné")),
    (DataGenerator.text(prefix="liné", return_bytes=True), DataGenerator.render(prefix="liné")),
])
def test_scrollable_listbox(inp, expected):
    lb = ScrollableListBox(MockUI(), inp)
    canvas = lb.render((SCREEN_WIDTH, SCREEN_HEIGHT))
    text = [bytes().join([t for at, cs, t in ln]) for ln in canvas.content()]
    assert text == expected


@pytest.mark.parametrize("inp,expected", [
    (DataGenerator.stream(), DataGenerator.render()),
    (DataGenerator.stream(return_bytes=True), DataGenerator.render()),
    (DataGenerator.stream(prefix="liné"), DataGenerator.render(prefix="liné")),
    (DataGenerator.stream(prefix="liné", return_bytes=True), DataGenerator.render(prefix="liné")),
])
def test_async_scrollable_listbox(inp, expected):
    ui = flexmock(refresh=lambda: None)
    lb = AsyncScrollableListBox(inp, ui)
    lb.thread.join()
    canvas = lb.render((SCREEN_WIDTH, SCREEN_HEIGHT))
    text = [bytes().join([t for at, cs, t in ln]) for ln in canvas.content()]
    w = "{:%d}" % SCREEN_WIDTH
    s = w.format("{}".format("No more logs."))
    expected[4] = s.encode("utf-8")
    assert text == expected


def test_table_random_data():
    rows = [ResponsiveRowWidget([get_random_text_widget(random.randint(2, 9)) for _ in range(5)])
            for _ in range(5)]
    table = ResponsiveTable(MockUI(), SimpleListWalker(rows))
    canvas = table.render((80, 20), focus=False)
    text = [bytes().join([t for at, cs, t in ln]) for ln in canvas.content()]
    logging.info("%r", text)
    assert len(text) == 20
    assert text[0].startswith(rows[0].original_widget.widget_list[0].text.encode("utf-8"))


def test_table_empty():
    rows = []
    table = ResponsiveTable(MockUI(), SimpleListWalker(rows))
    canvas = table.render((80, 20), focus=False)
    text = [bytes().join([t for at, cs, t in ln]) for ln in canvas.content()]
    assert len(text) == 20
    assert text[0] == b" " * 80


def test_assemble_rows_long_text():
    rows = [[get_random_text_widget(10),
             get_random_text_widget(300)] for _ in range(5)]
    assembled_rows = assemble_rows(rows, ignore_columns=[1])
    lb = WidgetBase(MockUI(), SimpleListWalker(assembled_rows))
    canvas = lb.render((80, 20), focus=False)
    text = [bytes().join([t for at, cs, t in ln]) for ln in canvas.content()]
    logging.info("%r", text)
    assert len(text) == 20
    first_col, second_col = text[0].split(b" ", 1)
    assert first_col == rows[0][0].text.encode("utf-8")
    assert rows[0][1].text.encode("utf-8").startswith(second_col)



# -*- coding: utf-8 -*-


import pytest

from sen.tui.commands.base import Command, Option, Commander, register_command, Argument


@register_command
class MyCommand(Command):
    name = "test"
    description = "test desc"
    options_definitions = [
        Option("test-opt", "test-opt desc", aliases=["x"])
    ]
    aliases = ["mango"]

    def run(self):
        return 42


@register_command
class MyCommand2(Command):
    name = "test2"
    description = "test desc 2"
    options_definitions = [
        Option("test-opt", "test-opt desc", aliases=["x"]),
        Option("test-opt2", "test-opt2 desc", default="banana"),
    ]
    arguments_definitions = [
        Argument("test-arg", "test-arg desc")
    ]

    def run(self):
        return 43


def test_command_class():
    c = Command()
    assert c.arguments is None
    c.process_args([])
    assert c.arguments is not None
    with pytest.raises(AttributeError):
        print(c.arguments.not_there)
    with pytest.raises(NotImplementedError):
        c.run()


def test_custom_command():
    c = MyCommand()
    c.process_args(["x"])
    assert c.arguments.test_opt is True
    assert c.run() == 42


def test_arg_and_opt():
    c = MyCommand2()
    c.process_args(["x", "y"])
    assert c.arguments.test_opt is True
    assert c.arguments.test_arg == "y"
    assert c.run() == 43


def test_default_arg():
    c = MyCommand2()
    c.process_args([])
    assert c.arguments.test_opt2 == "banana"
    assert c.run() == 43


def test_opt():
    c = MyCommand2()
    c.process_args(["y"])
    assert c.arguments.test_opt is None
    assert c.arguments.test_arg == "y"
    assert c.run() == 43


def test_commander():
    com = Commander(None, None)
    c = com.get_command("test test-opt")
    assert c.arguments.test_opt is True
    assert c.run() == 42


def test_command_aliases():
    com = Commander(None, None)
    c = com.get_command("mango x")
    assert c.arguments.test_opt is True
    assert c.run() == 42



import functools
import json
import logging
import datetime
import traceback
from operator import attrgetter

from sen.constants import ISO_DATETIME_PARSE_STRING
from sen.exceptions import (
    TerminateApplication, NotifyError, NotAvailableAnymore
)

import docker
import docker.errors

from sen.net import NetData
from sen.util import (
    calculate_cpu_percent, calculate_cpu_percent2, calculate_blkio_bytes,
    calculate_network_bytes, repeater,
    humanize_time,
    graceful_chain_get
)

logger = logging.getLogger(__name__)


DINOSAUR_TIME = datetime.datetime.fromordinal(1)


class ImageNameStruct(object):
    """
    stolen from atomic-reactor; thanks @mmilata!
    """
    def __init__(self, registry=None, namespace=None, repo=None, tag=None):
        self.registry = registry
        self.namespace = namespace
        self.repo = repo
        self.tag = tag

    @classmethod
    def parse(cls, image_name):
        result = cls()

        # registry.org/namespace/repo:tag
        s = image_name.split('/', 2)

        if len(s) == 2:
            if '.' in s[0] or ':' in s[0]:
                result.registry = s[0]
            else:
                result.namespace = s[0]
        elif len(s) == 3:
            result.registry = s[0]
            result.namespace = s[1]
        if result.namespace == 'library':
            # https://github.com/projectatomic/atomic-reactor/issues/45
            logger.debug("namespace 'library' -> ''")
            result.namespace = None
        result.repo = s[-1]

        try:
            result.repo, result.tag = result.repo.rsplit(':', 1)
        except ValueError:
            pass

        return result

    def to_str(self, registry=True, tag=True, explicit_tag=False,
               explicit_namespace=False):
        if self.repo is None:
            raise RuntimeError('No image repository specified')

        result = self.repo if self.repo != "<none>" else ""

        # don't display <none> junk
        if tag and self.tag and self.tag != "<none>":
            result = '{0}:{1}'.format(result, self.tag)
        elif tag and explicit_tag and self.tag != "<none>":
            result = '{0}:{1}'.format(result, 'latest')

        # don't display <none> junk
        if self.namespace:
            result = '{0}/{1}'.format(self.namespace, result)
        elif explicit_namespace:
            result = '{0}/{1}'.format('library', result)

        if registry and self.registry:
            result = '{0}/{1}'.format(self.registry, result)

        return result

    def __str__(self):
        return self.to_str(registry=True, tag=True)

    def __repr__(self):
        return "ImageName(image=%s)" % repr(self.to_str())

    def __eq__(self, other):
        return type(self) == type(other) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not self == other

    def __hash__(self):
        return hash(self.to_str())

    def copy(self):
        return ImageNameStruct(
            registry=self.registry,
            namespace=self.namespace,
            repo=self.repo,
            tag=self.tag)


def operation(fmt_str):
    def wrap(func):
        @functools.wraps(func)
        def wrapper(self, *args, **kwargs):
            pretty_message = ""

            before = datetime.datetime.now()
            response = func(self, *args, **kwargs)
            after = datetime.datetime.now()

            # we want milliseconds, not seconds
            command_took = (after - before).total_seconds() * 1000
            # # this line literally break zsh; wat?
            # logger.debug("%s(%s, %s) %s -> [%f ms]", func.__name__, args, kwargs, self,
            #              command_took)
            if fmt_str:
                pretty_message = fmt_str.format(object_type=getattr(self, "pretty_object_type", ""),
                                                object_short_name=getattr(self, "short_name", ""))
            return Operation(response, pretty_message=pretty_message, took=command_took)
        return wrapper
    return wrap


class Operation:
    """
    class for describing performed operation
    """

    def __init__(self, response, pretty_message="", took=None):
        self.response = response
        self.pretty_message = pretty_message
        self.took = took


class DockerObject:
    """
    Common base for images and containers
    """
    def __init__(self, data, docker_backend, object_id=None):
        self._id = object_id
        self._short_id = None
        self.data = data  # `client.containers` or `client.images`
        self.docker_backend = docker_backend
        self._created = None
        self._inspect = None
        self._names = None

    @property
    def d(self):
        """
        shortcut for instance of Docker client
        """
        return self.docker_backend.client

    @property
    def created_int(self):
        return self.data["Created"]

    @property
    def created(self):
        if self._created is None:
            self._created = datetime.datetime.fromtimestamp(self.data["Created"])
        return self._created

    def set_id(self):
        if self._id is None:
            try:
                self._id = self.data["Id"]
            except KeyError:
                raise RuntimeError("initial data not specified")

    @property
    def object_id(self):
        if self._id is None:
            self.set_id()
        return self._id

    @property
    def short_id(self):
        if self._short_id is None:
            self.set_id()
            if ":" in self._id:
                colon_index = self._id.index(":") + 1
                self._short_id = self._id[colon_index:][:12]
            else:
                self._short_id = self._id[:12]
        return self._short_id

    def display_time_created(self):
        return humanize_time(self.created)

    def display_formal_time_created(self):
        # http://tools.ietf.org/html/rfc2822.html#section-3.3
        return self.created.strftime("%d %b %Y, %H:%M:%S")

    def inspect(self, cached=True):
        raise NotImplementedError()

    def display_inspect(self):
        try:
            return json.dumps(self.inspect().response, indent=2)
        except docker.errors.NotFound:
            raise NotAvailableAnymore()

    @property
    def labels(self):
        labels = self.data["Labels"]
        return labels

    @property
    def natural_sort_value(self):
        return self.created

    def metadata_get(self, path, cached=True):
        """
        get metadata from inspect, specified by path

        :param path: list of str
        :param cached: bool, use cached version of inspect if available
        """
        try:
            value = graceful_chain_get(self.inspect(cached=cached).response, *path)
        except docker.errors.NotFound:
            logger.warning("object %s is not available anymore", self)
            raise NotAvailableAnymore()
        return value

    def refresh(self):
        """ refresh metadata """
        self.inspect(cached=False)

    def __eq__(self, other):
        return type(self) == type(other) and self._id == other._id

    def __ne__(self, other):
        return not self == other

    def __hash__(self):
        return hash(self._id)


class DockerImage(DockerObject):
    def __init__(self, data, docker_backend, object_id=None):
        super().__init__(data, docker_backend, object_id=object_id)
        self._unique_size = None
        self._total_size = None
        self._virtual_size = None
        self._shared_size = None

    @property
    def image_id(self):
        return self.object_id

    @property
    def parent_id(self):
        if self.data:
            return self.data.get("ParentId", None)
        else:
            return self.metadata_get(["Parent"])

    @property
    def pretty_object_type(self):
        return "Image"

    @property
    def parent_image(self):
        try:
            parent_id = self.parent_id
        except Exception as ex:
            logger.error("error while getting parent ID of image %s: %r", self, ex)
            logger.info(traceback.format_exc())
            raise
        if parent_id:
            return self.docker_backend.get_image_by_id(parent_id)
        else:
            return self.docker_backend.scratch_image

    @property
    def layers(self):
        """
        similar as parent images, except that it uses /history API endpoint
        :return:
        """
        # sample output:
        # {
        #     "Created": 1457116802,
        #     "Id": "sha256:507cb13a216097710f0d234668bf64a4c92949c573ba15eba13d05aad392fe04",
        #     "Size": 204692029,
        #     "Tags": [
        #         "docker.io/fedora:latest"
        #     ],
        #     "Comment": "",
        #     "CreatedBy": "/bin/sh -c #(nop) ADD file:bcb5e5c... in /"
        # }
        try:
            response = self.d.history(self.image_id)
        except docker.errors.NotFound:
            raise NotAvailableAnymore()

        layers = []
        for l in response:
            layer_id = l["Id"]
            if layer_id == "<missing>":
                layers.append(DockerImage(l, self.docker_backend))
            else:
                layers.append(self.docker_backend.get_image_by_id(layer_id))
        return layers

    @property
    def children(self):
        return self.docker_backend.get_images_for_parent(self)

    def get_next_sibling(self):
        imgs = self.parent_image.children
        if len(imgs) == 1:
            return None
        try:
            return imgs[imgs.index(self) + 1]
        except IndexError:
            return None

    def get_prev_sibling(self):
        imgs = self.parent_image.children
        if len(imgs) == 1:
            return None
        # 0 - 1 turns into -1 which turns into last element which creates cycle
        # which totally messes up whole tree
        prev_index = imgs.index(self) - 1
        if prev_index < 0:
            return None
        else:
            return imgs[prev_index]

    @property
    def command(self):
        cmd = self.metadata_get(["Config", "Cmd"])
        if cmd:
            return " ".join(cmd)
        return ""

    @property
    def container_command(self):
        # history item
        created_by = graceful_chain_get(self.data, "CreatedBy")
        if created_by:
            return created_by
        try:
            cmd = self.metadata_get(["ContainerConfig", "Cmd"])
        except NotAvailableAnymore:
            pass
        else:
            if cmd:
                return " ".join(cmd)
        return ""

    @property
    def comment(self):
        return self.metadata_get(["Comment"])

    @property
    def total_size(self):
        """
        Size of ALL layers in bytes

        :return: int or None
        """
        return self._total_size or self.data.get("Size", 0)

    @property
    def unique_size(self):
        """
        Size of ONLY this particular layer

        :return: int or None
        """
        self._virtual_size = self._virtual_size or \
                             graceful_chain_get(self.data, "VirtualSize", default=0)
        try:
            return self._virtual_size - self._shared_size
        except TypeError:
            return 0

    @property
    def shared_size(self):
        """
        I guess this is size of layers which are shared with some other image

        :return: int or None
        """
        return self._shared_size

    @property
    def names(self):
        if self._names is None:
            self._names = []
            if self.data is None:
                return self._names
            # RepoTags = image, Tags = output from `history` command, Tags can be None
            for t in self.data.get("RepoTags", self.data.get("Tags")) or []:
                image_name = ImageNameStruct.parse(t)
                if image_name.to_str():
                    self._names.append(image_name)
            # RepoDigests can be none
            for t in self.data.get("RepoDigests", []) or []:
                image_name = ImageNameStruct.parse(t)
                if image_name.to_str():
                    self._names.append(image_name)
            # sort by name length
            self._names.sort(key=lambda x: len(x.to_str()))
        return self._names

    @property
    def short_name(self):
        try:
            ins = self.names[0]
        except IndexError:
            return self.short_id
        if ins.repo == "<none>":
            return self.short_id
        return ins.to_str()

    def base_image(self):
        child_image = self
        while True:
            try:
                parent_image = self.docker_backend.get_image_by_id(child_image.parent_id)
            except Exception as ex:
                logger.warning("error while getting image by ID: %r", ex)
                parent_image = None
            if parent_image is None:
                try:
                    child_image = child_image.parent_image
                except Exception as ex:
                    logger.error("error while getting parent image for image %s: %r", self, ex)
                    return None
                if child_image is None:
                    return None
            else:
                return parent_image

    @operation("Inspect image {object_short_name}.")
    def inspect(self, cached=False):
        if self._inspect is None or cached is False:
            try:
                self._inspect = self.d.inspect_image(self.image_id)
            except docker.errors.NotFound:
                self._inspect = self._inspect or {}
        return self._inspect

    @operation("{object_type} {object_short_name} removed!")
    def remove(self, force=False):
        return self.d.remove_image(self.image_id, force=force)

    @operation("Tag of {object_type} {object_short_name} removed!")
    def remove_tag(self, tag):
        assert tag in self.names
        return self.d.remove_image(str(tag))

    def matches_search(self, s):
        return s in self.image_id or any([s in str(x) for x in self.names])

    def __str__(self):
        # it's dangerous to put many stuff here b/c most of the values are loaded dynamically
        # and it's trivial to go into nested exception madness
        return "image {}".format(self.short_id)

    def __repr__(self):
        return self.__str__()

    def containers(self):
        return self.docker_backend.get_containers_for_image(self.image_id)


class RootImage(DockerImage):
    """
    this is essentially "scratch" but you cannot inspect it anymore
    """
    def __init__(self, docker_backend):
        self.image_name = "scratch"
        super().__init__(None, docker_backend, object_id="")

    @property
    def parent_id(self):
        return None

    @property
    def parent_image(self):
        return None

    def get_next_sibling(self):
        return None

    def get_prev_sibling(self):
        return None

    @property
    def names(self):
        return [ImageNameStruct.parse(self.image_name)]

    def __str__(self):
        return self.image_name


class DockerContainer(DockerObject):
    """
    Container related logic
    """

    def __init__(self, data, docker_backend, object_id=None):
        super(DockerContainer, self).__init__(data, docker_backend, object_id)
        self.size_root_fs = None
        self.size_rw_fs = None

    def __str__(self):
        return "{} ({})".format(self.container_id, self.short_name)

    # properties

    @property
    def container_id(self):
        return self.object_id

    @property
    def names(self):
        if self._names is None:
            self._names = []
            for t in self.data.get("Names", []):
                self._names.append(t)
            # sort by name length
            self._names.sort(key=lambda x: len(x))
        return self._names

    @property
    def command(self):
        return self.data["Command"]

    @property
    def nice_status(self):
        return self.data["Status"]

    @property
    def simple_status(self):
        return self.metadata_get(["State", "Status"])

    @property
    def simple_status_cap(self):
        return self.simple_status.capitalize()

    @property
    def running(self):
        return self.metadata_get(["State", "Running"])

    @property
    def status_created(self):
        return self.simple_status == "created"

    @property
    def exit_code(self):
        return self.metadata_get(["State", "ExitCode"])

    @property
    def exited_well(self):
        return self.exit_code == 0

    @property
    def short_name(self):
        try:
            return self.names[0]
        except IndexError:
            return self.short_id

    @property
    def pretty_object_type(self):
        return "Container"

    @property
    def image_id(self):
        """ this container is created from image with id..."""
        try:
            # docker >= 1.9
            image_id = self.data["ImageID"]
        except KeyError:
            # docker <= 1.8
            image_id = self.metadata_get(["Image"])
        return image_id

    @property
    def image(self):
        return self.docker_backend.get_image_by_id(self.image_id)

    @property
    def ip_address(self):
        # docker == 1.10
        ip_address = self.metadata_get(["NetworkSettings", "IPAddress"])
        return ip_address

    @property
    def net(self):
        """
        get ACTIVE port mappings of a container

        :return: dict:
        {
            "host_port": "container_port"
        }
        """
        try:
            return NetData(self.inspect(cached=True).response)
        except docker.errors.NotFound:
            raise NotAvailableAnymore()

    @property
    def started_at(self):
        s = self.metadata_get(["State", "StartedAt"])
        if s:
            # python expects 6 digits in milliseconds, docker returns 9
            s = s[:26]
            if s == "0001-01-01T00:00:00Z":
                return DINOSAUR_TIME
            s = s.replace("Z", "0")
            try:
                started_at = datetime.datetime.strptime(s, ISO_DATETIME_PARSE_STRING)
            except ValueError as ex:
                logger.error("unable to parse datetime %s: %s", s, ex)
                return DINOSAUR_TIME
            return started_at

    @property
    def finished_at(self):
        f = self.metadata_get(["State", "FinishedAt"])
        if f:
            f = f[:26]
            if f == "0001-01-01T00:00:00Z":
                return DINOSAUR_TIME
            f = f.replace("Z", "0")
            try:
                finished_at = datetime.datetime.strptime(f, ISO_DATETIME_PARSE_STRING)
            except ValueError as ex:
                logger.error("unable to parse datetime %s: %s", f, ex)
                return DINOSAUR_TIME
            return finished_at

    @property
    def natural_sort_value(self):
        # docker acts weird: 'created' is provided as ordinal and is local time
        # 'started' and 'finished' are UTC as timestamp; it would be awesome to unite those b/c
        # atm these inconsistencies mess ordering
        try:
            # Nones are unsortable
            return max([x for x in
                        [self.started_at, self.finished_at, super().natural_sort_value]
                        if x])
        except NotAvailableAnymore:
            return super().natural_sort_value

    # methods

    def image_name(self):
        if self.image is not None:
            return self.image.short_name
        else:
            return self.image_id[:12]

    def matches_search(self, s):
        return s in self.container_id or \
               s in self.short_name
    # api calls

    @operation("Get resources statistics.")
    def stats(self):
        cpu_total = 0.0
        cpu_system = 0.0
        cpu_percent = 0.0
        for x in self.d.stats(self.container_id, decode=True, stream=True):
            blk_read, blk_write = calculate_blkio_bytes(x)
            net_r, net_w = calculate_network_bytes(x)
            mem_current = x["memory_stats"]["usage"]
            mem_total = x["memory_stats"]["limit"]

            try:
                cpu_percent, cpu_system, cpu_total = calculate_cpu_percent2(x, cpu_total, cpu_system)
            except KeyError as e:
                logger.error("error while getting new CPU stats: %r, falling back")
                cpu_percent = calculate_cpu_percent(x)

            r = {
                "cpu_percent": cpu_percent,
                "mem_current": mem_current,
                "mem_total": x["memory_stats"]["limit"],
                "mem_percent": (mem_current / mem_total) * 100.0,
                "blk_read": blk_read,
                "blk_write": blk_write,
                "net_rx": net_r,
                "net_tx": net_w,
            }
            yield r

    @operation("List processes in running container.")
    def top(self):
        """
        list of processes in a running container

        :return: None or list of dicts
        """
        # let's get resources from .stats()
        ps_args = "-eo pid,ppid,wchan,args"
        # returns {"Processes": [values], "Titles": [values]}
        # it's easier to play with list of dicts: [{"pid": 1, "ppid": 0}]
        try:
            response = self.d.top(self.container_id, ps_args=ps_args)
        except docker.errors.APIError as ex:
            logger.warning("error getting processes: %r", ex)
            return []
        # TODO: sort?
        logger.debug(json.dumps(response, indent=2))
        return [dict(zip(response["Titles"], process))
                for process in response["Processes"] or []]

    @operation("Inspect container {object_short_name}.")
    def inspect(self, cached=False):
        if cached is False or self._inspect is None:
            try:
                self._inspect = self.d.inspect_container(self.container_id)
            except docker.errors.NotFound:
                self._inspect = self._inspect or {}
        return self._inspect

    @operation("Logs of container {object_short_name} received.")
    def logs(self, follow=False, lines="all"):
        # when tail is set to all, it takes ages to populate widget
        # docker-py does `inspect` in the background
        try:
            logs_data = self.d.logs(self.container_id, stream=follow, tail=lines)
        except docker.errors.NotFound:
            return None
        else:
            return logs_data

    @operation("{object_type} {object_short_name} removed!")
    def remove(self, force=False):
        self.d.remove_container(self.container_id, force=force)

    @operation("{object_type} {object_short_name} started.")
    def start(self):
        self.d.start(self.container_id)

    @operation("{object_type} {object_short_name} stopped.")
    def stop(self):
        self.d.stop(self.container_id)

    @operation("{object_type} {object_short_name} restarted.")
    def restart(self):
        self.d.restart(self.container_id)

    @operation("{object_type} {object_short_name} killed.")
    def kill(self):
        self.d.kill(self.container_id)

    @operation("{object_type} {object_short_name} paused.")
    def pause(self):
        self.d.pause(self.container_id)

    @operation("{object_type} {object_short_name} unpaused.")
    def unpause(self):
        self.d.unpause(self.container_id)


class DockerBackend:
    """
    backend for docker
    """

    def __init__(self):
        self._containers = None
        self._images = None  # displayed images
        self._all_images = None  # docker images -a
        self._df = None

        kwargs = {"version": "auto"}
        kwargs.update(docker.utils.kwargs_from_env(assert_hostname=False))

        try:
            APIClientClass = docker.Client  # 1.x
        except AttributeError:
            APIClientClass = docker.APIClient  # 2.x

        try:
            self.client = APIClientClass(**kwargs)
        except docker.errors.DockerException as ex:
            raise TerminateApplication("can't establish connection to docker daemon: {0}".format(str(ex)))

        self.scratch_image = RootImage(self)

    # backend queries

    @operation("Get list of images.")
    def get_images(self, cached=True):
        if cached is False or self._images is None:
            logger.debug("doing images() query")
            self._images = {}
            images_response = repeater(self.client.images) or []
            for i in images_response:
                img = DockerImage(i, self)
                self._images[img.image_id] = img
            self._all_images = {}
            # FIXME: performance: do just all=True
            all_images_response = repeater(self.client.images, kwargs={"all": True}) or []
            for i in all_images_response:
                img = DockerImage(i, self)
                self._all_images[img.image_id] = img
        return list(self._images.values())

    @operation("Get list of containers.")
    def get_containers(self, cached=True, stopped=True):
        if cached is False or self._containers is None:
            logger.debug("doing containers() query")
            self._containers = {}
            containers_reponse = repeater(self.client.containers, kwargs={"all": stopped}) or []
            for c in containers_reponse:
                container = DockerContainer(c, self)
                self._containers[container.container_id] = container
        if not stopped:
            return [x for x in list(self._containers.values()) if x.running]
        return list(self._containers.values())

    @operation("Get disk usage.")
    def df(self, cached=True):
        if cached is False or self._df is None:
            logger.debug("getting disk-usage")
            # TODO: wrap in try/execpt
            self._df = self.client.df()
            # TODO: attach these to real containers and images
            # # since DOCKER API-1.25 (v.1.13.0)
            # df = self.client.df()
            # if 'Containers' in df:
            #     df_containers = df['Containers']
            containers_data = graceful_chain_get(self._df, "Containers")
            for c_data in containers_data:
                c = graceful_chain_get(self._containers, graceful_chain_get(c_data, "Id"))
                c.size_root_fs = graceful_chain_get(c_data, "SizeRootFs")
                c.size_rw_fs = graceful_chain_get(c_data, "SizeRw")
            images_data = graceful_chain_get(self._df, "Images")
            for i_data in images_data:
                i = graceful_chain_get(self._images, graceful_chain_get(i_data, "Id"))
                i._total_size = graceful_chain_get(i_data, "Size")
                i._shared_size = graceful_chain_get(i_data, "SharedSize")
                i._virtual_size = graceful_chain_get(i_data, "VirtualSize")
        return self._df

    def realtime_updates(self):
        event = it = None
        while True:
            if not it or not event:
                it = repeater(self.client.events, kwargs={"decode": True}, retries=5)
                if not it:
                    raise NotifyError("Unable to fetch realtime updates from docker engine.")

            event = repeater(next, args=(it, ), retries=2)  # likely an engine restart
            if not event:
                continue

            logger.debug("RT event: %s", event)

            yield event

            # try:
            #     # 1.10+
            #     is_container = event["Type"] == "container"
            # except KeyError:
            #     # event["from'] means it's a container
            #     is_container = "from" in event
            # if is_container:
            #     # inspect doesn't contain info about status and you can't query just one
            #     # container with containers()
            #     # let's do full-blown containers() query; it's not that expensive
            #     self.get_containers(cached=False)
            # else:
            #     # similar as ^
            #     # images() doesn't support query by ID
            #     # inspect doesn't contain info about repositories
            #     self.get_images(cached=False)
            # content, _, _ = self.filter(containers=True, images=True, stopped=True,
            #                             cached=True, sort_by_created=True)
            # yield content

    # service methods

    def get_image_by_id(self, image_id):
        return self._all_images.get(image_id)

    def get_images_for_parent(self, image):
        if not image:
            return []
        l = sorted([x for x in self._all_images.values() if x.parent_image == image], key=lambda x: x.created_int)
        return l

    def get_container_by_id(self, container_id):
        return self._containers.get(container_id)

    def get_containers_for_image(self, image_id):
        return [container for container in self._containers.values() if container.image_id == image_id]

    def filter(self, containers=True, images=True, stopped=True, cached=False, sort_by_created=True):
        """
        since django is so awesome, let's use their ORM API

        :return:
        """
        content = []
        containers_o = None
        images_o = None
        # return containers when containers=False and running=True
        if containers or not stopped:
            containers_o = self.get_containers(cached=cached, stopped=stopped)
            content += containers_o.response
        if images:
            images_o = self.get_images(cached=cached)
            content += images_o.response
        if sort_by_created:
            content.sort(key=attrgetter("natural_sort_value"), reverse=True)
        return content, containers_o, images_o



import logging

import urwid

from sen.docker_backend import RootImage
from sen.tui.chunks.container import ContainerOneLinerWidget
from sen.tui.chunks.image import LayerWidget
from sen.tui.views.base import View
from sen.tui.widgets.list.base import WidgetBase
from sen.tui.widgets.list.util import RowWidget
from sen.tui.widgets.table import assemble_rows
from sen.tui.widgets.util import SelectableText, get_map
from sen.util import humanize_bytes


logger = logging.getLogger(__name__)


class TagWidget(SelectableText):
    """
    so we can easily access image and tag
    """
    def __init__(self, docker_image, tag):
        self.docker_image = docker_image
        self.tag = tag
        super().__init__(str(self.tag))


class ImageInfoWidget(WidgetBase, View):
    """
    display info about image
    """
    def __init__(self, ui, docker_image):
        self.walker = urwid.SimpleFocusListWalker([])
        super().__init__(ui, self.walker)

        self.docker_image = docker_image

    def refresh(self):
        self.docker_image.refresh()
        self.walker.clear()
        self._basic_data()
        self._containers()
        self._image_names()
        self._layers()
        self._labels()
        self.set_focus(0)

    @property
    def focused_docker_object(self):
        # TODO: enable removing image names
        try:
            return self.focus.columns.widget_list[0].docker_container
        except AttributeError:
            try:
                return self.focus.columns.widget_list[0].docker_image
            except AttributeError:
                return None

    def _basic_data(self):
        data = [
            [SelectableText("Id", maps=get_map("main_list_green")),
             SelectableText(self.docker_image.image_id)],
            [SelectableText("Created", maps=get_map("main_list_green")),
             SelectableText("{0}, {1}".format(self.docker_image.display_formal_time_created(),
                                              self.docker_image.display_time_created()))],
            [SelectableText("Size", maps=get_map("main_list_green")),
             SelectableText(humanize_bytes(self.docker_image.total_size))],
        ]
        if self.docker_image.unique_size:
            data.append(
                [SelectableText("Unique Size", maps=get_map("main_list_green")),
                 SelectableText(humanize_bytes(self.docker_image.unique_size))])
        if self.docker_image.shared_size:
            data.append(
                [SelectableText("Shared Size", maps=get_map("main_list_green")),
                 SelectableText(humanize_bytes(self.docker_image.shared_size))])
        data.append([SelectableText("Command", maps=get_map("main_list_green")),
                     SelectableText(self.docker_image.container_command)])
        self.walker.extend(assemble_rows(data, ignore_columns=[1]))

    def _image_names(self):
        if not self.docker_image.names:
            return
        self.walker.append(RowWidget([SelectableText("")]))
        self.walker.append(RowWidget([SelectableText("Image Names", maps=get_map("main_list_white"))]))
        for n in self.docker_image.names:
            self.walker.append(RowWidget([TagWidget(self.docker_image, n)]))

    def _layers(self):
        self.walker.append(RowWidget([SelectableText("")]))
        self.walker.append(RowWidget([SelectableText("Layers", maps=get_map("main_list_white"))]))

        i = self.docker_image
        parent = i.parent_image
        layers = self.docker_image.layers
        index = 0

        if isinstance(parent, RootImage) and len(layers) > 0:  # pulled image, docker 1.10+
            for image in layers:
                self.walker.append(
                    RowWidget([LayerWidget(self.ui, image, index=index)])
                )
                index += 1
        else:
            self.walker.append(RowWidget([LayerWidget(self.ui, self.docker_image, index=index)]))
            while True:
                index += 1
                parent = i.parent_image
                if parent:
                    self.walker.append(RowWidget([LayerWidget(self.ui, parent, index=index)]))
                    i = parent
                else:
                    break

    def _labels(self):
        if not self.docker_image.labels:
            return []
        data = []
        self.walker.append(RowWidget([SelectableText("")]))
        self.walker.append(RowWidget([SelectableText("Labels", maps=get_map("main_list_white"))]))
        for label_key, label_value in self.docker_image.labels.items():
            data.append([SelectableText(label_key, maps=get_map("main_list_green")), SelectableText(label_value)])
        self.walker.extend(assemble_rows(data, ignore_columns=[1]))

    def _containers(self):
        if not self.docker_image.containers():
            return
        self.walker.append(RowWidget([SelectableText("")]))
        self.walker.append(RowWidget([SelectableText("Containers", maps=get_map("main_list_white"))]))
        for container in self.docker_image.containers():
            self.walker.append(RowWidget([ContainerOneLinerWidget(self.ui, container)]))



"""
TODO:
 * nicer list
 * summary
 * clickable items
 * enable deleting volumes
"""
import urwid

from sen.util import humanize_bytes, graceful_chain_get
from sen.tui.views.base import View
from sen.tui.widgets.list.util import SingleTextRow
from sen.tui.widgets.table import assemble_rows
from sen.tui.constants import MAIN_LIST_FOCUS
from sen.tui.widgets.util import SelectableText, get_map
from sen.tui.widgets.list.base import WidgetBase


class DfBufferView(WidgetBase, View):
    def __init__(self, ui, buffer):
        """

        :param ui:
        :param buffer: Buffer instance, display help about this buffer
        """
        self.ui = ui
        self.buffer = buffer
        self.walker = urwid.SimpleFocusListWalker([])
        super().__init__(ui, self.walker)

    def refresh(self, df=None, containers=None, images=None):
        content = []
        if df is None:
            content += [
                SingleTextRow("Data is being loaded, it may even take a couple minutes.",
                              maps={"normal": "main_list_white", "focus": MAIN_LIST_FOCUS}),
            ]
        else:
            if containers:
                content += [
                    SingleTextRow("Containers",
                                  maps={"normal": "main_list_white", "focus": MAIN_LIST_FOCUS}),
                    SingleTextRow("")
                ]
                containers_content = [[
                    SelectableText("Name", maps=get_map("main_list_lg")),
                    SelectableText("Image Size", maps=get_map("main_list_lg")),
                    SelectableText("Writable Layer Size", maps=get_map("main_list_lg")),
                ]]
                for c in containers:
                    containers_content.append(
                        [SelectableText(c.short_name),
                         SelectableText(humanize_bytes(c.size_root_fs or 0)),
                         SelectableText(humanize_bytes(c.size_rw_fs or 0)),
                    ])
                content.extend(assemble_rows(containers_content, dividechars=3))
                content += [
                    SingleTextRow("")
                ]
            if images:
                content += [
                    SingleTextRow("Images",
                                  maps={"normal": "main_list_white", "focus": MAIN_LIST_FOCUS}),
                    SingleTextRow("")
                ]
                images_content = [[
                    SelectableText("Name", maps=get_map("main_list_lg")),
                    SelectableText("Size", maps=get_map("main_list_lg")),
                    SelectableText("Shared Size", maps=get_map("main_list_lg")),
                    SelectableText("Unique Size", maps=get_map("main_list_lg"))
                ]]
                for i in images:
                    images_content.append([
                        SelectableText(i.short_name, maps=get_map("main_list_dg")),
                        SelectableText(
                            humanize_bytes(i.total_size or 0),
                            maps=get_map("main_list_dg")),
                        SelectableText(
                            humanize_bytes(i.shared_size or 0),
                            maps=get_map("main_list_dg")),
                        SelectableText(
                            humanize_bytes(i.unique_size or 0),
                            maps=get_map("main_list_dg"))
                    ])
                content.extend(assemble_rows(images_content, dividechars=3))
                content += [
                    SingleTextRow("")
                ]
            volumes = graceful_chain_get(df, "Volumes")
            if volumes:
                content += [
                    SingleTextRow("Volumes",
                                  maps={"normal": "main_list_white", "focus": MAIN_LIST_FOCUS}),
                    SingleTextRow("")
                ]
                volumes_content = [[
                    SelectableText("Name", maps=get_map("main_list_lg")),
                    SelectableText("Links", maps=get_map("main_list_lg")),
                    SelectableText("Size", maps=get_map("main_list_lg")),
                ]]
                for v in volumes:
                    v_name = graceful_chain_get(v, "Name", default="")
                    v_size = graceful_chain_get(v, "UsageData", "Size", default=0)
                    v_links = graceful_chain_get(v, "UsageData", "RefCount", default=0)
                    volumes_content.append([
                        SelectableText(v_name, maps=get_map("main_list_dg")),
                        SelectableText("%s" % v_links, maps=get_map("main_list_dg")),
                        SelectableText(
                            humanize_bytes(v_size),
                            maps=get_map("main_list_dg")),
                    ])
                content.extend(assemble_rows(volumes_content, dividechars=3))
        self.set_body(content)
        self.set_focus(0)



import logging
import re
import threading
import time

import urwid

from sen.exceptions import NotifyError, NotAvailableAnymore
from sen.tui.chunks.misc import get_row
from sen.tui.widgets.list.util import (
    get_operation_notify_widget, ResponsiveRowWidget
)
from sen.tui.widgets.table import ResponsiveTable
from sen.util import graceful_chain_get

logger = logging.getLogger(__name__)


class MainLineWidget(ResponsiveRowWidget):
    def __init__(self, docker_object):
        self.docker_object = docker_object
        super().__init__(get_row(docker_object))

    def matches_search(self, s):
        return self.docker_object.matches_search(s)

    def __repr__(self):
        return "{}({})".format(self.__class__.__name__, self.docker_object)

    def __str__(self):
        return "{}".format(self.docker_object)


class MainListBox(ResponsiveTable):
    def __init__(self, ui, docker_backend):
        self.d = docker_backend
        super(MainListBox, self).__init__(ui, urwid.SimpleFocusListWalker([]))

        # urwid is not thread safe
        self.refresh_lock = threading.Lock()

        # realtime lock
        self.realtime_lock = threading.Lock()

        self.stop_realtime_events = threading.Event()

    def refresh(self, query=None):
        """
        refresh listing, also apply filters
        :return:
        """
        logger.info("refresh listing")
        focus_on_top = len(self.body) == 0  # focus if empty
        with self.refresh_lock:
            self.query(query_string=query)
        if focus_on_top:
            try:
                self.set_focus(0)
            except IndexError:
                pass

    def process_realtime_event(self, event):
        with self.realtime_lock:
            if self.stop_realtime_events.is_set():
                logger.info("received docker event when this functionality is disabled")
                return
        delayed_events = ["pause", "unpause"]
        ev_status = graceful_chain_get(event, "status")
        if ev_status in delayed_events:
            # https://github.com/TomasTomecek/sen/issues/143
            # tl;dr dockerd does not tell us when the container is in pause/unpause state
            # it sends the event when the container is being paused
            time.sleep(1)
        self.refresh(query=self.filter_query)

    def filter(self, s, widgets_to_filter=None):
        self.refresh(query=s)

    def toggle_realtime_events(self):
        with self.realtime_lock:
            if self.stop_realtime_events.is_set():
                self.stop_realtime_events.clear()
                self.ui.notify_message("Enabling live updates from docker.")
            else:
                self.stop_realtime_events.set()
                self.ui.notify_message("Disabling live updates from docker.")
        self.ui.reload_footer()

    def query(self, query_string=""):
        """
        query and display, also apply filters

        :param query_string: str
        :return: None
        """

        def query_notify(operation):
            w = get_operation_notify_widget(operation, display_always=False)
            if w:
                self.ui.notify_widget(w)

        if query_string is not None:
            self.filter_query = query_string.strip()

        # FIXME: this could be part of filter command since it's command line
        backend_query = {
            "cached": False,
            "containers": True,
            "images": True,
        }

        def containers():
            backend_query["containers"] = True
            backend_query["images"] = not backend_query["images"]
            backend_query["cached"] = True

        def images():
            backend_query["containers"] = not backend_query["containers"]
            backend_query["images"] = True
            backend_query["cached"] = True

        def running():
            backend_query["stopped"] = False
            backend_query["cached"] = True
            backend_query["images"] = False

        query_conf = [
            {
                "query_keys": ["t", "type"],
                "query_values": ["c", "container", "containers"],
                "callback": containers
            }, {
                "query_keys": ["t", "type"],
                "query_values": ["i", "images", "images"],
                "callback": images
            }, {
                "query_keys": ["s", "state"],
                "query_values": ["r", "running"],
                "callback": running
            },
        ]
        query_list = re.split(r"[\s,]", self.filter_query)
        unprocessed = []
        for query_str in query_list:
            if not query_str:
                continue
            # process here x=y queries and pass rest to parent filter()
            try:
                query_key, query_value = query_str.split("=", 1)
            except ValueError:
                unprocessed.append(query_str)
            else:
                logger.debug("looking up query key %r and query value %r", query_key, query_value)
                for c in query_conf:
                    if query_key in c["query_keys"] and query_value in c["query_values"]:
                        c["callback"]()
                        break
                else:
                    raise NotifyError("Invalid query string: %r", query_str)

        widgets = []
        logger.debug("doing query %s", backend_query)
        query, c_op, i_op = self.d.filter(**backend_query)

        for o in query:
            try:
                line = MainLineWidget(o)
            except NotAvailableAnymore:
                continue
            widgets.append(line)
        if unprocessed:
            new_query = " ".join(unprocessed)
            logger.debug("doing parent query for unprocessed string: %r", new_query)
            super().filter(new_query, widgets_to_filter=widgets)
        else:
            self.set_body(widgets)
            self.ro_content = widgets

        query_notify(i_op)
        query_notify(c_op)

    def status_bar(self):
        columns_list = []

        def add_subwidget(markup, color_attr=None):
            if color_attr is None:
                w = urwid.AttrMap(urwid.Text(markup), "status_text")
            else:
                w = urwid.AttrMap(urwid.Text(markup), color_attr)
            columns_list.append((len(markup), w))

        add_subwidget("Images: ")
        images_count = len(self.d.get_images(cached=True).response)
        if images_count < 10:
            add_subwidget(str(images_count), "status_text_green")
        elif images_count < 50:
            add_subwidget(str(images_count), "status_text_yellow")
        else:
            add_subwidget(str(images_count), "status_text_orange")

        add_subwidget(", Containers: ")
        containers_count = len(self.d.get_containers(cached=True, stopped=True).response)
        if containers_count < 5:
            add_subwidget(str(containers_count), "status_text_green")
        elif containers_count < 30:
            add_subwidget(str(containers_count), "status_text_yellow")
        elif containers_count < 100:
            add_subwidget(str(containers_count), "status_text_orange")
        else:
            add_subwidget(str(containers_count), "status_text_red")

        add_subwidget(", Running: ")
        running_containers = self.d.get_containers(cached=True, stopped=False).response
        running_containers_n = len(running_containers)
        add_subwidget(str(running_containers_n),
                      "status_text_green" if running_containers_n > 0 else "status_text")

        with self.realtime_lock:
            if self.stop_realtime_events.is_set():
                add_subwidget(", Live updates are disabled")

        parent_cols = super().status_bar()
        if parent_cols:
            add_subwidget(", ")
        return columns_list + parent_cols



"""
commands related to displaying stuff

FIXME: some of these commands duplicate stuff in ui
"""

import logging

from sen.docker_backend import DockerImage, DockerContainer
from sen.exceptions import NotifyError
from sen.tui.buffer import (
    TreeBuffer, HelpBuffer,
    ImageInfoBuffer, ContainerInfoBuffer,
    MainListBuffer
)
from sen.tui.commands.base import FrontendCommand, register_command


logger = logging.getLogger(__name__)


@register_command
class DisplayLayersTreeCommand(FrontendCommand):
    name = "display-layers-tree"

    def run(self):
        self.ui.add_and_display_buffer(TreeBuffer(self.docker_object, self.ui))


@register_command
class DisplayHelpCommand(FrontendCommand):
    name = "display-help"

    def run(self):
        self.ui.add_and_display_buffer(HelpBuffer(self.ui, self.buffer))


@register_command
class DisplayInfoBufferCommand(FrontendCommand):
    name = "display-info"

    def run(self):
        # TODO: needs better abstraction, backend object should be tied together with frontend
        # object via a new Class
        if isinstance(self.docker_object, DockerImage):
            buffer_class = ImageInfoBuffer
        elif isinstance(self.docker_object, DockerContainer):
            buffer_class = ContainerInfoBuffer
        else:
            self.ui.notify_message("Can't display info for '%s'" % self.docker_object,
                                   level="error")
            logger.error("unable to display info buffer for %r", self.docker_object)
            return
        try:
            # FIXME: this try/except block should be in upper frame
            self.ui.add_and_display_buffer(buffer_class(self.docker_object, self.ui))
        except NotifyError as ex:
            self.ui.notify_message(str(ex), level="error")
            logger.error(repr(ex))


@register_command
class DisplayListingCommand(FrontendCommand):
    name = "display-listing"

    def run(self):
        b = MainListBuffer(self.ui, self.docker_backend)
        self.ui.add_and_display_buffer(b, redraw=True)



import logging

import urwid

from sen.tui.widgets.list.base import WidgetBase
from sen.tui.widgets.list.util import RowWidget

logger = logging.getLogger(__name__)


def calculate_max_cols_length(table, size):
    """
    :param table: list of lists:

    [["row 1 column 1", "row 1 column 2"],
     ["row 2 column 1", "row 2 column 2"]]

    each item consists of instance of urwid.Text

    :returns dict, {index: width}
    """
    max_cols_lengths = {}

    for row in table:
        col_index = 0
        for idx, widget in enumerate(row.widgets):
            l = widget.pack((size[0], ))[0]
            max_cols_lengths[idx] = max(max_cols_lengths.get(idx, 0), l)
            col_index += 1

    max_cols_lengths.setdefault(0, 1)  # in case table is empty
    return max_cols_lengths


def assemble_rows(data, max_allowed_lengths=None, dividechars=1,
                  ignore_columns=None):
    """
    :param data: list of lists:
    [["row 1 column 1", "row 1 column 2"],
     ["row 2 column 1", "row 2 column 2"]]
    each item consists of instance of urwid.Text

    :param max_allowed_lengths: dict:
        {col_index: maximum_allowed_length}
    :param ignore_columns: list of ints, indexes which should not be calculated
    """
    rows = []
    max_lengths = {}
    ignore_columns = ignore_columns or []

    # shitty performance, here we go
    # it would be way better to do a single double loop and provide mutable variable
    # FIXME: merge this code with calculate() from above
    for row in data:
        col_index = 0
        for widget in row:
            if col_index in ignore_columns:
                continue
            l = len(widget.text)
            if max_allowed_lengths:
                if col_index in max_allowed_lengths and max_allowed_lengths[col_index] < l:
                    # l is bigger then what is allowed
                    l = max_allowed_lengths[col_index]

            max_lengths.setdefault(col_index, l)
            max_lengths[col_index] = max(l, max_lengths[col_index])
            col_index += 1

    for row in data:
        row_widgets = []
        for idx, item in enumerate(row):
            if idx in ignore_columns:
                row_widgets.append(item)
            else:
                row_widgets.append((max_lengths[idx], item))
        rows.append(
            RowWidget(row_widgets, dividechars=dividechars)
        )

    return rows


class ResponsiveTable(WidgetBase):
    def __init__(self, ui, walker, headers=None, dividechars=1, responsive=True):
        """
        :param walker: list of ResponsiveRow instances
        :param headers: list of widgets which should be displayed as headers
        """
        self.dividechars = dividechars
        self.responsive = responsive

        super().__init__(ui, walker)

    def render(self, size, focus=False):
        screen_width = size[0]

        # max text length for each column -- table
        min_col_lengths = calculate_max_cols_length(self.body, size)

        # compute maximal column width -- looks nicer
        max_col_width = int(screen_width / len(min_col_lengths)) - self.dividechars
        for k, v in min_col_lengths.items():
            min_col_lengths[k] = min(min_col_lengths[k], max_col_width)

        columns_occupy = sum(min_col_lengths.values())

        if self.responsive:
            # responsibility! spread the columns
            to_spread = screen_width - len(min_col_lengths) * self.dividechars - columns_occupy
            if to_spread:
                spread_remaining = to_spread
                weights = {}  # longer cols will get more space
                sum_weights = 0
                sorted_by_length = sorted(min_col_lengths.items(), key=lambda x: x[1])
                longest_col = sorted_by_length[-1][0]
                for weight, (idx, length) in enumerate(sorted_by_length):
                    weights[idx] = weight + 1  # first weight is 1; 0 doesn't add anything
                    sum_weights += weight + 1
                for k, v in min_col_lengths.items():
                    expansion = int(to_spread * weights[k] / sum_weights)
                    min_col_lengths[k] += expansion
                    spread_remaining -= expansion
                # add remaining to the longest col
                min_col_lengths[longest_col] += spread_remaining

        for row in self.body:
            row.contents[:] = [
                (w, (urwid.GIVEN, min_col_lengths[idx], is_box))
                for idx, (w, (_, _, is_box)) in enumerate(row.contents)
            ]

        return super().render(size, focus=focus)





MAIN_LIST_FOCUS = "main_list_focus"

STATUS_BG = "#06a"
STATUS_BG_FOCUS = "#08d"

# name, fg, bg, mono, fg_h, bg_h
PALLETE = [
    (MAIN_LIST_FOCUS, 'default', 'brown', "default", "white", "#060"),  # a60
    ('main_list_lg', 'light gray', 'default', "default", "g100", "default"),
    ('main_list_dg', 'dark gray', 'default', "default", "g78", "default"),
    ('main_list_ddg', 'dark gray', 'default', "default", "g56", "default"),
    ('main_list_white', 'white', 'default', "default", "white", "default"),
    ('main_list_green', 'dark green', 'default', "default", "#0f0", "default"),
    ('main_list_yellow', 'brown', 'default', "default", "#ff0", "default"),
    ('main_list_orange', 'light red', 'default', "default", "#fa0", "default"),
    ('main_list_red', 'dark red', 'default', "default", "#f00", "default"),
    ('image_names', 'light magenta', 'default', "default", "#F0F", "default"),
    ('status_box', 'default', 'black', "default", "g100", STATUS_BG),
    ('status_box_focus', 'default', 'black', "default", "white", STATUS_BG_FOCUS),
    ('status', 'default', 'default', "default", "default", STATUS_BG),
    ('status_text', 'default', 'default', "default", "g100", STATUS_BG),
    ('status_text_green', 'default', 'default', "default", "#0f0", STATUS_BG),
    ('status_text_yellow', 'default', 'default', "default", "#ff0", STATUS_BG),
    ('status_text_orange', 'default', 'default', "default", "#f80", STATUS_BG),
    ('status_text_red', 'default', 'default', "default", "#f66", STATUS_BG),
    ('notif_error', "white", 'dark red', "default", "white", "#f00",),
    ('notif_info', 'white', 'default', "default", "g100", "default"),
    ('notif_important', 'white', 'default', "default", "white", "default"),
    ('notif_text_green', 'white', 'default', "white", "#0f0", "default"),
    ('notif_text_yellow', 'white', 'default', "white", "#ff0", "default"),
    ('notif_text_orange', 'white', 'default', "white", "#f80", "default"),
    ('notif_text_red', 'white', 'default', "white", "#f66", "default"),
    ('tree', 'dark green', 'default', "default", "dark green", "default"),

    ('graph_bg', "default", 'default', "default", "default", "default"),

    ('graph_lines_cpu', "default", 'default', "default", "default", "#d63"),
    ('graph_lines_cpu_tips', "default", 'default', "default", "#d63", "default"),
    ('graph_lines_cpu_legend', "default", 'default', "default", "#f96", "default"),

    ('graph_lines_mem', "default", 'default', "default", "default", "#39f"),
    ('graph_lines_mem_tips', "default", 'default', "default", "#39f", "default"),
    ('graph_lines_mem_legend', "default", 'default', "default", "#6af", "default"),

    ('graph_lines_blkio_r', "default", 'default', "default", "default", "#9b0"),
    ('graph_lines_blkio_r_tips', "default", 'default', "default", "#9b0", "default"),
    ('graph_lines_blkio_r_legend', "default", 'default', "default", "#cf0", "default"),
    ('graph_lines_blkio_w', "default", 'default', "default", "default", "#b90"),
    ('graph_lines_blkio_w_tips', "default", 'default', "default", "#b90", "default"),
    ('graph_lines_blkio_w_legend', "default", 'default', "default", "#fc0", "default"),

    ('graph_lines_net_r', "default", 'default', "default", "default", "#3ca"),
    ('graph_lines_net_r_tips', "default", 'default', "default", "#3ca", "default"),
    ('graph_lines_net_r_legend', "default", 'default', "default", "#6fc", "default"),
    ('graph_lines_net_w', "default", 'default', "default", "default", "#3ac"),
    ('graph_lines_net_w_tips', "default", 'default', "default", "#3ac", "default"),
    ('graph_lines_net_w_legend', "default", 'default', "default", "#6cf", "default"),
]

STATUS_BAR_REFRESH_SECONDS = 5
CLEAR_NOTIF_BAR_MESSAGE_IN = 5



import logging
import threading

import urwid

from sen.tui.constants import MAIN_LIST_FOCUS

logger = logging.getLogger(__name__)


def get_map(defult="main_list_dg"):
    return {"normal": defult, "focus": MAIN_LIST_FOCUS}


class AdHocAttrMap(urwid.AttrMap):
    """
    Ad-hoc attr map change

    taken from https://github.com/pazz/alot/
    """
    def __init__(self, w, maps, init_map='normal'):
        self.maps = maps
        urwid.AttrMap.__init__(self, w, maps[init_map])
        if isinstance(w, urwid.Text):
            self.attrs = [x[0] for x in self.original_widget.get_text()[1]]

    def set_map(self, attrstring):
        attr_map = {None: self.maps[attrstring]}

        # for urwid.Text only: do hovering for all markups in the widget
        if isinstance(self.original_widget, urwid.Text):
            if attrstring == "normal":
                for a in self.attrs:
                    attr_map[self.maps["focus"]] = a
            elif attrstring == "focus":
                for a in self.attrs:
                    attr_map[a] = self.maps["focus"]
        self.set_attr_map(attr_map)


class ColorTextMixin:
    @property
    def text(self):
        return self.original_widget.text

    @text.setter
    def text(self, value):
        self.original_widget.set_text(value)

    def keypress(self, size, key):
        """ get rid of tback: `AttributeError: 'Text' object has no attribute 'keypress'` """
        return key


class ColorText(ColorTextMixin, urwid.AttrMap):
    def __init__(self, text, color):
        super().__init__(urwid.Text(text, align="left", wrap="clip"), color)


class SelectableText(ColorTextMixin, AdHocAttrMap):
    def __init__(self, text, maps=None):
        maps = maps or get_map()
        super().__init__(urwid.Text(text, align="left", wrap="clip"), maps)


class ThreadSafeFrame(urwid.Frame):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.update_lock = threading.Lock()

    def set_body(self, body):
        with self.update_lock:
            return super().set_body(body=body)

    def set_footer(self, footer):
        with self.update_lock:
            return super().set_footer(footer=footer)

    def set_header(self, header):
        with self.update_lock:
            return super().set_header(header=header)

    def render(self, size, focus=False):
        with self.update_lock:
            return super().render(size=size, focus=focus)


class UnselectableListBox(urwid.ListBox):
    _selectable = False



"""
Application specific code.
"""

import logging
import threading

from sen.exceptions import NotifyError
from sen.tui.commands.base import Commander, SameThreadPriority
from sen.tui.commands.display import DisplayListingCommand
from sen.tui.ui import get_app_in_loop
from sen.tui.constants import PALLETE
from sen.docker_backend import DockerBackend


logger = logging.getLogger(__name__)


class Application:
    def __init__(self, yolo=False):
        self.d = DockerBackend()

        self.loop, self.ui = get_app_in_loop(PALLETE)

        self.ui.yolo = yolo

        self.ui.commander = Commander(self.ui, self.d)

        self.rt_thread = threading.Thread(target=self.realtime_updates, daemon=True)
        self.rt_thread.start()

    def run(self):
        self.ui.run_command(DisplayListingCommand.name, queue=SameThreadPriority())
        self.loop.run()

    def realtime_updates(self):
        """
        fetch realtime events from docker and pass them to buffers

        :return: None
        """
        # TODO: make this available for every buffer
        logger.info("starting receiving events from docker")
        it = self.d.realtime_updates()
        while True:
            try:
                event = next(it)
            except NotifyError as ex:
                self.ui.notify_message("error when receiving realtime events from docker: %s" % ex,
                                       level="error")
                return
            # FIXME: we should pass events to all buffers
            # ATM the buffers can't be rendered since they are not displayed
            # and hence traceback like this: ListBoxError("Listbox contents too short! ...
            logger.debug("pass event to current buffer %s", self.ui.current_buffer)
            try:
                self.ui.current_buffer.process_realtime_event(event)
            except Exception as ex:
                # swallow any exc
                logger.error("error while processing runtime event: %r", ex)



"""
Base class for views, the API.
"""


class View:
    # TODO: views should have API for searching and filtering
    #       there should be a map between widgets and their searchable content
    def refresh(self):
        raise NotImplementedError()



from sen.docker_backend import DockerBackend
from sen.util import calculate_cpu_percent2, calculate_cpu_percent
from .real import image_data, mock, container_data


def test_images_call():
    mock()
    b = DockerBackend()
    operation = b.get_images(cached=False)
    images_response = operation.response
    assert len(images_response) == 1
    assert images_response[0].image_id == image_data[0]["Id"]
    assert images_response[0].short_name == image_data[0]["RepoTags"][0]
    assert images_response[0].parent_id == image_data[0]["ParentId"]
    assert images_response[0].created_int == image_data[0]["Created"]
    assert [str(x) for x in images_response[0].names] == image_data[0]["RepoTags"]


def test_containers_call():
    mock()
    b = DockerBackend()
    operation = b.get_containers(cached=False)
    containers_response = operation.response
    assert len(containers_response) == 1
    assert containers_response[0].container_id == container_data["Id"]
    assert containers_response[0].names == container_data["Names"]
    assert containers_response[0].short_name == container_data["Names"][0]
    assert containers_response[0].created_int == container_data["Created"]
    assert containers_response[0].command == container_data["Command"]
    assert containers_response[0].nice_status == container_data["Status"]
    assert containers_response[0].image_id == container_data["ImageID"]
    # assert containers_response[0].image_name()


def test_short_id():
    mock()
    b = DockerBackend()
    operation = b.get_images()
    images_response = operation.response
    assert images_response[0].short_id == image_data[0]["Id"][image_data[0]["Id"].index(":")+1:][:12]


def test_top():
    pass


def test_stats():
    mock()
    b = DockerBackend()
    c = b.get_containers()
    c0 = c.response.pop()
    operation = c0.stats()
    stats_stream = operation.response
    assert next(stats_stream)

    for x in c0.d.stats('x', decode=True, stream=True):
        calculate_cpu_percent(x)

    t = 0.0
    s = 0.0
    for x in c0.d.stats('x', decode=True, stream=True):
        calculate_cpu_percent(x)
        _, s, t = calculate_cpu_percent2(x, t, s)



import logging

from sen.constants import FALLBACK_LOG_PATH

__version__ = '0.6.2'


def set_logging(name="sen", level=logging.DEBUG, path=FALLBACK_LOG_PATH):
    logger = logging.getLogger(name)
    # do not propagate logs from logger 'sen' to root logger (as they could be accidentally
    # displayed in terminal)
    logger.propagate = False
    logger.setLevel(level)

    handler = logging.FileHandler(path)
    handler.setLevel(logging.DEBUG)
    # handler = logging.StreamHandler(sys.stderr)

    formatter = logging.Formatter(
        '%(asctime)s.%(msecs).03d %(filename)-17s %(levelname)-6s %(message)s', '%H:%M:%S')
    handler.setFormatter(formatter)
    logger.addHandler(handler)



"""
This is a framework for terminal interfaces built on top of urwid.Frame.

It must NOT contain any application specific code.
"""
import logging
import threading
from concurrent.futures.thread import ThreadPoolExecutor

import urwid

from sen.exceptions import NotifyError
from sen.tui.commands.base import (
    FrontendPriority, BackendPriority,
    SameThreadPriority, KeyNotMapped
)
from sen.tui.constants import CLEAR_NOTIF_BAR_MESSAGE_IN
from sen.tui.widgets.util import ThreadSafeFrame
from sen.util import log_traceback, OrderedSet


logger = logging.getLogger(__name__)


class ConcurrencyMixin:
    def __init__(self):
        # worker for long-running tasks - requests
        self.worker = ThreadPoolExecutor(max_workers=4)
        # worker for quick ui operations
        self.ui_worker = ThreadPoolExecutor(max_workers=2)

    @staticmethod
    def _run(worker, f, *args, **kwargs):
        # TODO: do another wrapper to wrap notify exceptions and show them
        f = log_traceback(f)
        worker.submit(f, *args, **kwargs)

    def run_in_background(self, task, *args, **kwargs):
        logger.info("running task %r(%s, %s) in background", task, args, kwargs)
        self._run(self.worker, task, *args, **kwargs)

    def run_quickly_in_background(self, task, *args, **kwargs):
        logger.info("running a quick task %r(%s, %s) in background", task, args, kwargs)
        self._run(self.ui_worker, task, *args, **kwargs)


class UI(ThreadSafeFrame, ConcurrencyMixin):
    """
    handles all UI-specific code
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # widget -> message or None
        self.widget_message_dict = {}
        # message -> widget
        self.message_widget_dict = {}
        self.status_bar = None
        self.prompt_bar = None

        # lock when managing notifications:
        #  * when accessing self.notification_*
        #  * when accessing widgets
        # and most importantly, remember, locking is voodoo
        self.notifications_lock = threading.RLock()

        # populated when loop and UI are instantiated
        self.loop = None
        self.commander = None

        self.buffers = []
        self.buffer_movement_history = OrderedSet()

        self.main_list_buffer = None  # singleton

        self.current_buffer = None

    def refresh(self):
        self.loop.refresh()

    def quit(self):
        """
        This could be called from another thread, so let's do this via alarm
        """
        def q(*args):
            raise urwid.ExitMainLoop()
        self.worker.shutdown(wait=False)
        self.ui_worker.shutdown(wait=False)
        self.loop.set_alarm_in(0, q)

    # FIXME: move these to separate mixin
    def _set_main_widget(self, widget, redraw):
        """
        add provided widget to widget list and display it

        :param widget:
        :return:
        """
        self.set_body(widget)
        self.reload_footer()
        if redraw:
            logger.debug("redraw main widget")
            self.refresh()

    def display_buffer(self, buffer, redraw=True):
        """
        display provided buffer

        :param buffer: Buffer
        :return:
        """
        logger.debug("display buffer %r", buffer)
        self.buffer_movement_history.append(buffer)
        self.current_buffer = buffer
        self._set_main_widget(buffer.widget, redraw=redraw)

    def add_and_display_buffer(self, buffer, redraw=True):
        """
        add provided buffer to buffer list and display it

        :param buffer:
        :return:
        """
        # FIXME: some buffers have arguments, do a proper comparison -- override __eq__
        if buffer not in self.buffers:
            logger.debug("adding new buffer {!r}".format(buffer))
            self.buffers.append(buffer)
        self.display_buffer(buffer, redraw=redraw)

    def pick_and_display_buffer(self, i):
        """
        pick i-th buffer from list and display it

        :param i: int
        :return: None
        """
        if len(self.buffers) == 1:
            # we don't need to display anything
            # listing is already displayed
            return
        else:
            try:
                self.display_buffer(self.buffers[i])
            except IndexError:
                # i > len
                self.display_buffer(self.buffers[0])

    @property
    def current_buffer_index(self):
        return self.buffers.index(self.current_buffer)

    def remove_current_buffer(self, close_if_no_buffer=False):
        if len(self.buffers) == 1 and not close_if_no_buffer:
            return
        self.buffers.remove(self.current_buffer)
        self.buffer_movement_history.remove(self.current_buffer)
        self.current_buffer.destroy()
        if len(self.buffers) > 0:
            self.display_buffer(self.buffer_movement_history[-1], True)
        return len(self.buffers)

    def reload_footer(self, refresh=True, rebuild_statusbar=True):
        logger.debug("reload footer")
        footer = list(self.widget_message_dict.keys())
        if self.prompt_bar:
            footer.append(self.prompt_bar)
        else:
            if rebuild_statusbar or self.status_bar is None:
                self.status_bar = self.build_statusbar()
            footer.append(self.status_bar)
        # logger.debug(footer)
        self.set_footer(urwid.Pile(footer))
        if refresh:
            self.loop.refresh()

    def build_statusbar(self):
        """construct and return statusbar widget"""
        if self.prompt_bar:
            logger.info("prompt is active, won't build status bar")
            return
        try:
            left_widgets = self.current_buffer.build_status_bar() or []
        except AttributeError:
            left_widgets = []
        text_list = []
        # FIXME: this code should be placed in buffer
        # TODO: display current active worker threads
        for idx, buffer in enumerate(self.buffers):
            #  #1 [I] fedora #2 [L]
            fmt = "#{idx} [{name}]"
            markup = fmt.format(idx=idx, name=buffer.display_name)
            text_list.append((
                "status_box_focus" if buffer == self.current_buffer else "status_box",
                markup,
            ))
            text_list.append(" ")
        text_list = text_list[:-1]

        if text_list:
            buffer_text = urwid.Text(text_list, align="right")
        else:
            buffer_text = urwid.Text("", align="right")
        columns = urwid.Columns(left_widgets + [buffer_text])
        return urwid.AttrMap(columns, "status")

    def remove_notification_message(self, message):
        logger.debug("requested remove of message %r from notif bar", message)
        with self.notifications_lock:
            try:
                w = self.message_widget_dict[message]
            except KeyError:
                logger.warning("there is no notification %r displayed: %s",
                               message, self.message_widget_dict)
                return
            else:
                logger.debug("remove widget %r from new pile", w)
                del self.widget_message_dict[w]
                del self.message_widget_dict[message]
        self.reload_footer(rebuild_statusbar=False)

    def remove_widget(self, widget, message=None):
        logger.debug("remove widget %r from notif bar", widget)
        with self.notifications_lock:
            try:
                del self.widget_message_dict[widget]
            except KeyError:
                logger.info("widget %s was already removed", widget)
                return
            if message:
                del self.message_widget_dict[message]
        self.reload_footer(rebuild_statusbar=False)

    def notify_message(self, message, level="info", clear_if_dupl=True,
                       clear_in=CLEAR_NOTIF_BAR_MESSAGE_IN):
        """
        :param message, str
        :param level: str, {info, error}
        :param clear_if_dupl: bool, if True, don't display the notification again
        :param clear_in: seconds, remove the notificantion after some time

        opens notification popup.
        """
        with self.notifications_lock:
            if clear_if_dupl and message in self.message_widget_dict.keys():
                logger.debug("notification %r is already displayed", message)
                return
            logger.debug("display notification %r", message)
            widget = urwid.AttrMap(urwid.Text(message), "notif_{}".format(level))
        return self.notify_widget(widget, message=message, clear_in=clear_in)

    def notify_widget(self, widget, message=None, clear_in=CLEAR_NOTIF_BAR_MESSAGE_IN):
        """
        opens notification popup.

        :param widget: instance of Widget, widget to display
        :param message: str, message to remove from list of notifications
        :param clear_in: int, time seconds when notification should be removed
        """

        @log_traceback
        def clear_notification(*args, **kwargs):
            # the point here is the log_traceback
            self.remove_widget(widget, message=message)

        if not widget:
            return

        logger.debug("display notification widget %s", widget)

        with self.notifications_lock:
            self.widget_message_dict[widget] = message
            if message:
                self.message_widget_dict[message] = widget

        self.reload_footer(rebuild_statusbar=False)
        self.loop.set_alarm_in(clear_in, clear_notification)

        return widget

    def run_command(self, command_input, queue=None, **kwargs):
        kwargs["buffer"] = self.current_buffer
        command = self.commander.get_command(command_input, **kwargs)
        if command is None:
            return
        if queue is None:
            queue = command.priority
        if isinstance(queue, FrontendPriority):
            self.run_quickly_in_background(command.run)
        elif isinstance(queue, BackendPriority):
            self.run_in_background(command.run)
        elif isinstance(queue, SameThreadPriority):
            logger.info("running command %s", command)
            try:
                command.run()
            except NotifyError as ex:
                self.notify_message(str(ex), level="error")
                logger.error(repr(ex))
        else:
            raise RuntimeError("command %s doesn't have any priority: %s %s" %
                               (command_input, command.priority, FrontendPriority))

    def run_command_by_key(self, key, size, **kwargs):
        command_input = self.commander.get_command_input_by_key(key)
        self.run_command(command_input, size=size, **kwargs)

    def keypress(self, size, key):
        logger.debug("%s keypress %r", self.__class__.__name__, key)

        # we should pass the key to header, body, footer first so it's consumed in e.g. statusbar
        key = super().keypress(size, key)
        if key is None:
            logger.info("key was consumed by frame components")
            return

        logger.info("key was not consumed by frame components")

        focused_docker_object = None
        selected_widget = getattr(self.current_buffer, "widget", None)
        if selected_widget:
            focused_docker_object = getattr(self.current_buffer.widget, "focused_docker_object", None)
            logger.debug("focused docker object is %s", focused_docker_object)
        try:
            self.run_command_by_key(
                key,
                docker_object=focused_docker_object,
                size=size
            )
        except KeyNotMapped as ex:
            super_class = ThreadSafeFrame
            logger.debug("calling: %s.keypress(%s, %s)", super_class, size, key)
            # TODO: up/down doesn't do anything if len(lines) < screen height, that's confusing
            key = super_class.keypress(self, size, key)
            if key:
                self.notify_message(str(ex), level="error")
            logger.debug("was key handled? %s", "yes" if key is None else "no")
            return key
        return


class ThreadSafeLoop(urwid.MainLoop):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.refresh_lock = threading.RLock()

    def entering_idle(self):
        with self.refresh_lock:
            return super().entering_idle()

    def refresh(self):
        """
        explicitely refresh user interface; useful when changing widgets dynamically
        """
        logger.debug("refresh user interface")
        try:
            with self.refresh_lock:
                self.draw_screen()
        except AssertionError:
            logger.warning("application is not running")
            pass


def get_app_in_loop(pallete):
    screen = urwid.raw_display.Screen()
    screen.set_terminal_properties(256)
    screen.register_palette(pallete)

    ui = UI(urwid.SolidFill())
    decorated_ui = urwid.AttrMap(ui, "root")
    loop = ThreadSafeLoop(decorated_ui, screen=screen, event_loop=urwid.AsyncioEventLoop(),
                          handle_mouse=False)
    ui.loop = loop

    return loop, ui



class NotAvailableAnymore(Exception):
    """
    This object was available but isn't anymore.
    """


class NotifyError(Exception):
    """
    There was an error, we should notify user.
    """


class TerminateApplication(Exception):
    """
    Close application gracefully.
    """



import logging

from sen.docker_backend import DockerContainer, RootImage
from sen.exceptions import NotifyError
from sen.tui.commands.base import Command
from sen.tui.views.disk_usage import DfBufferView
from sen.tui.views.help import HelpBufferView, HelpCommandView
from sen.tui.views.main import MainListBox
from sen.tui.views.image_info import ImageInfoWidget
from sen.tui.views.container_info import ContainerInfoView
from sen.tui.widgets.list.common import AsyncScrollableListBox, ScrollableListBox
from sen.tui.widgets.list.util import get_operation_notify_widget
from sen.tui.widgets.tree import ImageTree


logger = logging.getLogger(__name__)


class Buffer:
    """
    base buffer class
    """
    name = None  # unique identifier
    description = None  # for help
    display_name = None  # display in status bar
    widget = None  # display this in main frame

    # global keybinds which will be available in every buffer
    global_keybinds = {
        # navigation
        "home": "navigate-top",
        "gg": "navigate-top",
        "end": "navigate-bottom",
        "G": "navigate-bottom",
        "down": "navigate-down",
        "j": "navigate-down",
        "up": "navigate-up",
        "k": "navigate-up",
        "ctrl d": "navigate-downwards",
        "ctrl u": "navigate-upwards",

        # UI
        ":": "prompt",
        "/": "prompt prompt-text=\"\" initial-text=\"/\"",
        "n": "search-next",
        "N": "search-previous",
        "f4": "prompt initial-text=\"filter \"",
        "x": "kill-buffer",
        "q": "kill-buffer quit-if-no-buffer",
        "ctrl i": "select-next-buffer",
        "ctrl o": "select-previous-buffer",
        "h": "help",
        "?": "help",
        "f5": "layers",
    }

    # buffer specific keybinds
    keybinds = {}

    def __init__(self):
        logger.debug("creating buffer %r", self)
        self._keybinds = None  # cache
        self.refresh()

    def __repr__(self):
        return "{}(name={!r}, widget={!r})".format(
            self.__class__.__name__, self.display_name, self.widget)

    def destroy(self):
        destroy_method = getattr(self.widget, "destroy", None)
        if destroy_method:
            destroy_method()

    def find_previous(self, s=None):
        logger.debug("searching next %r in %r", s, self.__class__.__name__)
        try:
            self.widget.find_previous(s)
        except AttributeError as ex:
            logger.debug(repr(ex))
            raise NotifyError("Can't search in this buffer.")

    def find_next(self, s=None):
        logger.debug("searching next %r in %r", s, self.__class__.__name__)
        try:
            self.widget.find_next(s)
        except AttributeError as ex:
            logger.debug(repr(ex))
            raise NotifyError("Can't search in this buffer.")

    def build_status_bar(self):
        status_bar = getattr(self.widget, "status_bar", None)
        if status_bar:
            return status_bar()

    def filter(self, s):
        logger.debug("filter widget %r with query %r", self.widget, s)
        self.widget.filter(s)

    def get_keybinds(self):
        if self._keybinds is None:
            self._keybinds = {}
            self._keybinds.update(self.global_keybinds)
            self._keybinds.update(self.keybinds)
        return self._keybinds

    def refresh(self):
        refresh_func = getattr(self.widget, "refresh", None)
        if refresh_func:
            logger.info("refreshing widget %s", self.widget)
            refresh_func()

    def process_realtime_event(self, event):
        logger.info("buffer %s doesn't process realtime events", self)
        return


class ImageInfoBuffer(Buffer):
    description = "Dashboard for information about selected image.\n" + \
                  "You can run command `df` to get more detailed info about disk usage."
    keybinds = {
        "enter": "display-info",
        "d": "rm",
        "i": "inspect",
        "@": "refresh",
    }

    def __init__(self, docker_image, ui):
        """
        :param docker_image:
        :param ui: ui object so we refresh
        """
        if isinstance(docker_image, RootImage):
            raise NotifyError("Image \"scratch\" doesn't provide any more information.")
        if docker_image.image_id == "<missing>":
            raise NotifyError("This image (layer) is not available due to changes in docker-1.10 "
                              "image representation.")
        self.docker_image = docker_image
        self.display_name = docker_image.short_name
        self.widget = ImageInfoWidget(ui, docker_image)
        super().__init__()

    def process_realtime_event(self, event):
        if event.get("id", None) == self.docker_image.object_id:
            self.widget.refresh()


class ContainerInfoBuffer(Buffer):
    description = "Detailed info about selected container presented in a slick dashboard."
    keybinds = {
        "enter": "display-info",
        "@": "refresh",
        "i": "inspect",
    }

    def __init__(self, docker_container, ui):
        """
        :param docker_container:
        :param ui: ui object so we refresh
        """
        self.docker_container = docker_container
        self.display_name = docker_container.short_name
        self.widget = ContainerInfoView(ui, docker_container)
        super().__init__()

    def process_realtime_event(self, event):
        action = event.get("Action", None)
        if action == "top":
            return
        if event.get("id", None) == self.docker_container.object_id:
            self.widget.refresh()


class TreeBuffer(Buffer):
    display_name = "Layers"
    description = "Tree view of all layers available on your docker engine."
    keybinds = {
        "enter": "display-info",
    }

    def __init__(self, ui, docker_backend):
        self.widget = ImageTree(ui, docker_backend)
        super().__init__()


class MainListBuffer(Buffer):
    display_name = "Listing"
    description = "List of all known docker images and containers display in a single list"
    keybinds = {
        "d": "rm",
        "D": "rm -f",
        "s": "start",
        "t": "stop",
        "r": "restart",
        "X": "kill",
        "p": "pause",
        "u": "unpause",
        "enter": "display-info",
        "b": "open-browser",
        "l": "logs",
        "f": "logs -f",
        "i": "inspect",
        "!": "toggle-live-updates",  # TODO: rfe: move to global so this affects every buffer
        "@": "refresh",  # FIXME: move to global and refactor & rewrite
    }

    def __init__(self, ui, docker_backend):
        self.ui = ui
        self.widget = MainListBox(ui, docker_backend)
        super().__init__()

    def process_realtime_event(self, event):
        self.widget.process_realtime_event(event)


class LogsBuffer(Buffer):
    description = "Display logs of selected container."
    display_name = "Logs "

    def __init__(self, ui, docker_object, follow=False):
        """

        :param docker_object: container to display logs
        :param ui: ui object so we can refresh
        """
        self.display_name += "({})".format(docker_object.short_name)
        if isinstance(docker_object, DockerContainer):
            try:
                pre_message = "Getting logs for container {}...".format(docker_object.short_name)
                ui.notify_message(pre_message)
                if follow:
                    # FIXME: this is a bit race-y -- we might lose some logs with this approach
                    operation = docker_object.logs(follow=follow, lines=0)
                    static_data = docker_object.logs(follow=False).response
                    self.widget = AsyncScrollableListBox(operation.response, ui, static_data=static_data)
                else:
                    operation = docker_object.logs(follow=follow)
                    self.widget = ScrollableListBox(ui, operation.response)
                ui.remove_notification_message(pre_message)
                ui.notify_widget(get_operation_notify_widget(operation, display_always=False))
            except Exception as ex:
                # FIXME: let's catch 404 and print that container doesn't exist
                #        instead of printing ugly HTTP error
                raise NotifyError("Error getting logs for container %s: %r" % (docker_object, ex))
        else:
            raise NotifyError("Only containers have logs.")
        super().__init__()


class InspectBuffer(Buffer):
    display_name = "Inspect "
    description = "Display all the information docker knows about selected object: " + \
                  "same output as `docker inspect`."

    def __init__(self, ui, docker_object):
        """

        :param docker_object: object to inspect
        """
        self.docker_object = docker_object
        self.ui = ui
        self.widget = None
        self.display_name += docker_object.short_name
        super().__init__()

    def refresh(self):
        inspect_data = self.docker_object.display_inspect()
        self.widget = ScrollableListBox(self.ui, inspect_data)

    def process_realtime_event(self, event):
        if event.get("id", None) == self.docker_object.object_id:
            self.ui.notify_message("Docker object changed, refreshing.")
            focus = self.widget.get_focus()[1]
            self.widget.set_text(self.docker_object.display_inspect())
            self.widget.set_focus(focus)


class HelpBuffer(Buffer):
    # TODO: apply this interface to other buffers: create views
    description = "Show information about currently displayed buffer and " + \
                  "what keybindings are available there"
    display_name = "Help"

    def __init__(self, ui, inp):
        """
        display buffer with more info about object 'inp'

        :param ui: UI instance
        :param inp: Buffer, Command instance
        """
        self.ui = ui
        if isinstance(inp, Buffer):
            self.display_name += "({})".format(inp.display_name)
            self.widget = HelpBufferView(ui, inp, self.global_keybinds)
        elif isinstance(inp, Command):
            self.display_name += "({})".format(inp.name)
            self.widget = HelpCommandView(ui, inp)

        super().__init__()


class DfBuffer(Buffer):
    description = "Show information about how much disk space container, images and volumes take."
    display_name = "Disk Usage"

    def __init__(self, ui):
        """
        :param ui: UI instance
        """
        self.ui = ui
        self.widget = DfBufferView(ui, self)

        super().__init__()

    def refresh(self, df=None, containers=None, images=None):
        self.widget.refresh(df=df, containers=containers, images=images)



"""
Image specific chunks.
"""

import logging

from sen.util import humanize_bytes

from sen.docker_backend import RootImage
from sen.tui.widgets.list.util import get_time_attr_map
from sen.tui.widgets.util import SelectableText, get_map


logger = logging.getLogger(__name__)


class LayerWidget(SelectableText):
    def __init__(self, ui, docker_image, index=None):
        self.ui = ui
        self.docker_image = docker_image
        label = []
        logger.debug("creating layer widget for %s", docker_image)
        if index is not None:
            separator = "└─"
            if index == 0:
                label = [separator]
            else:
                label = [2 * index * " " + separator]
        super().__init__(label + get_basic_image_markup(docker_image, with_size=True))


def get_detailed_image_row(docker_image):
    row = []
    image_id = SelectableText(docker_image.short_id, maps=get_map())
    row.append(image_id)

    command = SelectableText(docker_image.command, maps=get_map(defult="main_list_ddg"))
    row.append(command)

    base_image = docker_image.base_image()
    base_image_text = ""
    if base_image:
        base_image_text = base_image.short_name
    base_image_w = SelectableText(base_image_text, maps=get_map())
    row.append(base_image_w)

    time = SelectableText(docker_image.display_time_created(),
                          maps=get_time_attr_map(docker_image.created))
    row.append(time)

    image_names_markup = get_image_names_markup(docker_image)
    # urwid.Text([]) tracebacks
    if image_names_markup:
        image_names = SelectableText(image_names_markup)
    else:
        image_names = SelectableText("")
    row.append(image_names)

    return row


def get_image_names_markup(docker_image):
    text_markup = []
    for n in docker_image.names:
        if n.registry:
            text_markup.append(("main_list_dg", n.registry + "/"))
        if n.namespace and n.repo:
            text_markup.append(("main_list_lg", n.namespace + "/" + n.repo))
        else:
            if n.repo == "<none>":
                text_markup.append(("main_list_dg", n.repo))
            else:
                text_markup.append(("main_list_lg", n.repo))
        if n.tag:
            if n.tag not in ["<none>", "latest"]:
                text_markup.append(("main_list_dg", ":" + n.tag))
        text_markup.append(("main_list_dg", ", "))
    text_markup = text_markup[:-1]
    return text_markup


def get_basic_image_markup(docker_image, with_size=False):
    if isinstance(docker_image, RootImage):
        return [str(docker_image)]

    text_markup = [docker_image.short_id]

    if docker_image.names:
        text_markup.append(" ")
        text_markup.append(("main_list_lg", docker_image.names[0].to_str()))

    c = docker_image.container_command or docker_image.comment
    if c:
        text_markup.append(" ")
        text_markup.append(("main_list_ddg", c))

    if with_size:
        text_markup.append(" ")
        text_markup.append(("main_list_ddg", "(%s)" % humanize_bytes(docker_image.total_size)))

    return text_markup



"""
these commands may take long to finish, they are usually being run against external service (e.g.
docker engine)
"""
import logging
import webbrowser

from sen.tui.buffer import LogsBuffer, InspectBuffer, DfBuffer
from sen.tui.commands.base import BackendCommand, register_command, Option
from sen.tui.widgets.list.util import get_operation_notify_widget
from sen.docker_backend import DockerContainer


logger = logging.getLogger(__name__)


class OperationCommand(BackendCommand):
    def do(self, fn_name, pre_message=None, notif_level="info", **kwargs):
        pre_message = pre_message or self.pre_info_message.format(
            container_name=self.docker_object.short_name)
        self.ui.notify_message(pre_message)
        try:
            operation = getattr(self.docker_object, fn_name)(**kwargs)
        except AttributeError:
            log_txt = "you can't {} {}".format(fn_name, self.docker_object)
            logger.error(log_txt)
            notif_txt = "You can't {} {} {!r}.".format(
                fn_name,
                self.docker_object.pretty_object_type.lower(),
                self.docker_object.short_name)
            self.ui.notify_message(notif_txt, level="error")
        except Exception as ex:
            self.ui.notify_message(str(ex), level="error")
            raise
        else:
            self.ui.remove_notification_message(pre_message)
            self.ui.notify_widget(
                get_operation_notify_widget(operation, notif_level=notif_level)
            )


class MatchingOperationCommand(OperationCommand):
    """ this is just a shortcut """
    def run(self):
        self.do(self.name)


@register_command
class RemoveCommand(OperationCommand):
    name = "rm"
    description = "remove provided object, image or container"
    options_definitions = [Option("force",
                                  "Force removal of the selected object.",
                                  default=False, aliases=["-f", "f"]),
                           Option("yes",
                                  "Don't ask before removing.",
                                  default=False, aliases=["-y"])]

    def run(self):
        logger.debug("remove %s force=%s yes=%s", self.docker_object, self.arguments.force,
                     self.arguments.yes)
        if not self.arguments.yes and not self.ui.yolo:
            logger.debug("we need confirmation from user")
            cmd = "rm -y" if not self.arguments.force else "rm -y -f"
            self.ui.run_command(
                "prompt prompt-text=\"You are about to remove %s %s, type enter to continue: \" "
                "initial-text=\"%s\"" % (
                    self.docker_object.pretty_object_type.lower(),
                    self.docker_object.short_name,
                    cmd
                ),
                docker_object=self.docker_object
            )
            return

        if self.arguments.force:
            self.ui.notify_message("Removing forcibly!", level="important")
        self.do("remove",
                pre_message="Removing {} {}...".format(
                    self.docker_object.pretty_object_type.lower(),
                    self.docker_object.short_name),
                notif_level="important",
                force=self.arguments.force)


@register_command
class StartContainerCommand(MatchingOperationCommand):
    name = "start"
    description = "start a container"
    pre_info_message = "Starting container {container_name}..."


@register_command
class StopContainerCommand(MatchingOperationCommand):
    name = "stop"
    description = "stop a container"
    pre_info_message = "Stopping container {container_name}..."


@register_command
class RestartContainerCommand(MatchingOperationCommand):
    name = "restart"
    description = "restart a container"
    pre_info_message = "Restarting container {container_name}..."


@register_command
class KillContainerCommand(MatchingOperationCommand):
    name = "kill"
    description = "kill a container"
    pre_info_message = "Killing container {container_name}..."


@register_command
class PauseContainerCommand(MatchingOperationCommand):
    name = "pause"
    description = "pause a container"
    pre_info_message = "Pausing container {container_name}..."


@register_command
class UnpauseContainerCommand(MatchingOperationCommand):
    name = "unpause"
    description = "unpause a container"
    pre_info_message = "Unpausing container {container_name}..."


@register_command
class LogsCommand(BackendCommand):
    name = "logs"
    description = "display logs of a container"
    options_definitions = [Option("follow", "Follow logs.", default=False, aliases=["-f", "f"])]

    def run(self):
        self.ui.add_and_display_buffer(
            LogsBuffer(self.ui, self.docker_object, follow=self.arguments.follow))


@register_command
class InspectCommand(BackendCommand):
    name = "inspect"
    description = "inspect provided object, a container or an image"

    def run(self):
        if not self.docker_object:
            self.ui.notify_message("No docker object specified.", level="error")
            return
        self.ui.add_and_display_buffer(InspectBuffer(self.ui, self.docker_object))


@register_command
class DfCommand(BackendCommand):
    name = "df"
    description = "show disk usage"

    def run(self):
        b = DfBuffer(self.ui)
        self.ui.add_and_display_buffer(b)
        df = self.docker_backend.df()
        b.refresh(df=df.response,
                  containers=self.docker_backend.get_containers(cached=True, stopped=True).response,
                  images=self.docker_backend.get_images(cached=True).response)


@register_command
class OpenPortsInBrowser(BackendCommand):
    name = "open-browser"
    # TODO: user should be able to specify port and ip: by hitting keybind in container info view
    description = "open exposed port in a browser"

    def run(self):
        if not isinstance(self.docker_object, DockerContainer):
            self.ui.notify_message("No docker container specified.", level="error")
            return

        if not self.docker_object.running:
            self.ui.notify_message("Container is not running - no ports are available.",
                                   level="error")
            return

        ports = list(self.docker_object.net.ports.keys())
        ips = self.docker_object.net.ips

        logger.debug("ports = %s, ips = %s", ports, ips)

        if not ports:
            self.ui.notify_message(
                "Container %r doesn't expose any ports." % self.docker_object.short_name,
                level="error"
            )
            return

        url = "http://{}:{}".format(ips[list(ips.keys())[0]]["ip_address4"], ports[0])
        logger.info("opening %s in browser", url)
        webbrowser.open(url)



import logging
import re
import threading
import traceback

import urwid

from sen.tui.widgets.list.base import WidgetBase
from sen.util import _ensure_unicode


logger = logging.getLogger(__name__)


# def translate_asci_sequence(s):
#     # FIXME: not finished
#     translation_map = {
#         "34": "dark blue"
#     }
#     return translation_map.get(s, "")


def strip_from_ansi_esc_sequences(text):
    """
    find ANSI escape sequences in text and remove them

    :param text: str
    :return: list, should be passed to ListBox
    """
    # esc[ + values + control character
    # h, l, p commands are complicated, let's ignore them
    seq_regex = r"\x1b\[[0-9;]*[mKJusDCBAfH]"
    regex = re.compile(seq_regex)
    start = 0
    response = ""
    for match in regex.finditer(text):
        end = match.start()
        response += text[start:end]

        start = match.end()
    response += text[start:len(text)]
    return response


# def colorize_text(text):
#     """
#     finds ANSI color escapes in text and transforms them to urwid
#
#     :param text: str
#     :return: list, should be passed to ListBox
#     """
#     # FIXME: not finished
#     response = []
#     # http://ascii-table.com/ansi-escape-sequences.php
#     regex_pattern = r"(?:\x1b\[(\d+)?(?:;(\d+))*m)([^\x1b]+)"  # [%d;%d;...m
#     regex = re.compile(regex_pattern, re.UNICODE)
#     for match in regex.finditer(text):
#         groups = match.groups()
#         t = groups[-1]
#         color_specs = groups[:-1]
#         urwid_spec = translate_asci_sequence(color_specs)
#         if urwid_spec:
#             item = (urwid.AttrSpec(urwid_spec, "main_list_dg"), t)
#         else:
#             item = t
#         item = urwid.AttrMap(urwid.Text(t, align="left", wrap="any"), "main_list_dg", "main_list_white")
#         response.append(item)
#     return response


class ScrollableListBox(WidgetBase):
    def __init__(self, ui, text, focus_bottom=True):
        self.walker = urwid.SimpleFocusListWalker([])
        super().__init__(ui, self.walker)
        self.set_text(text)
        if focus_bottom:
            try:
                self.set_focus(len(self.walker) - 2)
            except IndexError:
                pass

    def set_text(self, text):
        self.walker.clear()
        text = _ensure_unicode(text)
        # logger.debug(repr(text))
        text = strip_from_ansi_esc_sequences(text)
        list_of_texts = text.split("\n")
        self.walker[:] = [
            urwid.AttrMap(urwid.Text(t.rstrip(), align="left", wrap="any"), "main_list_dg", "main_list_white")
            for t in list_of_texts
        ]


class AsyncScrollableListBox(WidgetBase):
    def __init__(self, generator, ui, static_data=None):
        self.log_texts = []
        if static_data:
            static_data = _ensure_unicode(static_data).split("\n")
            for d in static_data:
                log_entry = d.rstrip()
                if log_entry:
                    self.log_texts.append(urwid.Text(("main_list_dg", log_entry),
                                                     align="left", wrap="any"))
        walker = urwid.SimpleFocusListWalker(self.log_texts)
        super(AsyncScrollableListBox, self).__init__(ui, walker)
        walker.set_focus(len(walker) - 1)

        def fetch_logs():
            line_w = urwid.AttrMap(
                urwid.Text("", align="left", wrap="any"), "main_list_dg", "main_list_white"
            )
            walker.append(line_w)

            while True:
                try:
                    line = next(generator)
                except StopIteration:
                    logger.info("no more logs")
                    line_w = urwid.AttrMap(
                        urwid.Text("No more logs.", align="left", wrap="any"),
                        "main_list_dg", "main_list_white"
                    )
                    walker.append(line_w)
                    walker.set_focus(len(walker) - 1)
                    break
                except Exception as ex:
                    logger.error(traceback.format_exc())
                    ui.notify_message("Error while fetching logs: %s", ex)
                    break
                line = _ensure_unicode(line)
                if self.stop.is_set():
                    break
                if self.filter_query:
                    if self.filter_query not in line:
                        continue
                line_w.original_widget.set_text(line_w.original_widget.text + line.rstrip("\r\n"))
                if line.endswith("\n"):
                    walker.set_focus(len(walker) - 1)
                    line_w = urwid.AttrMap(
                        urwid.Text("", align="left", wrap="any"), "main_list_dg", "main_list_white"
                    )
                    walker.append(line_w)
                ui.refresh()

        self.stop = threading.Event()
        self.thread = threading.Thread(target=fetch_logs, daemon=True)
        self.thread.start()

    def destroy(self):
        self.stop.set()




"""
This suite should test whether sen is capable of running in concurrent high load environment
"""
import random
import threading

from sen.tui.ui import UI
from sen.tui.widgets.list.base import WidgetBase

import urwid

from .utils import get_random_text_widget, get_random_text


class MockUI:
    buffers = []

    def refresh(self):
        pass

    def set_alarm_in(self, *args, **kwargs):
        pass


def test_main_frame():
    lower_bound = 5
    upper_bound = 10
    list_count = 50

    def loop_skeleton(greater_f, less_f, l, l_b, u_b, l_c):
        # we could do while True, but then we need to clean the threads
        for _ in range(100):
            if len(l) > l_c:
                for _ in range(random.randint(l_b, u_b)):
                    greater_f()
            else:
                for _ in range(random.randint(l_b, u_b)):
                    less_f()

    body_widgets = [get_random_text_widget() for _ in range(100)]
    ui = MockUI()
    frame = WidgetBase(ui, urwid.SimpleFocusListWalker(body_widgets))

    def add_and_remove_random():
        widgets = []

        def less_f():
            if random.randint(1, 2) % 2:
                w = get_random_text_widget()
                frame.notify_widget(w)
                widgets.append(w)
            else:
                widgets.append(frame.notify_message(get_random_text()))

        def greater_f():
            w = random.choice(widgets)
            frame.remove_widget(w)
            widgets.remove(w)
        loop_skeleton(greater_f, less_f, widgets, lower_bound, upper_bound, list_count)

    def change_body():

        def less_f():
            body_widgets.insert(0, get_random_text_widget())

        def greater_f():
            body_widgets.remove(random.choice(body_widgets))
        loop_skeleton(greater_f, less_f, body_widgets, lower_bound, upper_bound, list_count)

    nt = threading.Thread(target=add_and_remove_random, daemon=True)
    bt = threading.Thread(target=change_body, daemon=True)
    nt.start()
    bt.start()
    for x in range(50):
        frame.render((70, 70))
    nt.join()
    bt.join()



# -*- coding: utf-8 -*-
import logging
import time
from concurrent.futures.thread import ThreadPoolExecutor
from datetime import datetime, timedelta

from sen.tui.widgets.list.common import strip_from_ansi_esc_sequences
from sen.util import _ensure_unicode, log_traceback, repeater, humanize_time, \
    OrderedSet

import pytest


@pytest.mark.parametrize("inp,expected", [
    ("a", "a"),
    (b"a", "a"),
    ("\u2606", "☆"),
    (b'\xe2\x98\x86', "☆"),
])
def test_ensure_unicode(inp, expected):
    assert _ensure_unicode(inp) == expected


def test_log_traceback(caplog):
    @log_traceback
    def f():
        raise Exception()
    caplog.set_level(logging.DEBUG)
    f()
    assert caplog.records[0].message.endswith(" is about to be started")
    assert caplog.records[1].message.startswith("Traceback")
    assert caplog.records[1].message.endswith("Exception\n")


def test_log_traceback_without_tb(caplog):
    @log_traceback
    def f():
        pass
    caplog.set_level(logging.DEBUG)
    f()
    assert caplog.records[0].message.endswith(" is about to be started")
    assert caplog.records[1].message.endswith(" finished")


def test_log_traceback_threaded(caplog):
    @log_traceback
    def f():
        raise Exception()
    caplog.set_level(logging.DEBUG)

    e = ThreadPoolExecutor(max_workers=1)
    f = e.submit(f)
    while f.running():
        time.sleep(0.1)

    assert caplog.records[0].message.endswith(" is about to be started")
    assert caplog.records[1].message.startswith("Traceback")
    assert caplog.records[1].message.endswith("Exception\n")


# def test_log_vars_from_tback(caplog):
#     a = 1
#     b = None
#     c = []
#     try:
#         raise Exception()
#     except Exception:
#         log_vars_from_tback(4)
#
#     def has_similar_message(msg):
#         for log_entry in caplog.records():
#             if msg in log_entry.message:
#                 return True
#         return False
#
#     assert has_similar_message("c = []")
#     assert has_similar_message("b = None")
#     assert has_similar_message("a = 1")


def test_repeater():
    def g(l, item=1):
        l.append(1)
        return l[item]

    assert repeater(g, args=([], )) == 1
    with pytest.raises(Exception):
        repeater(f, args=([], ), kwargs={"item": 10}) == 1


@pytest.mark.parametrize("inp,expected", [
    (timedelta(seconds=1), "now"),
    (timedelta(seconds=2), "2 seconds ago"),
    (timedelta(seconds=59), "59 seconds ago"),
    (timedelta(minutes=1), "1 minute ago"),
    (timedelta(minutes=1, seconds=1), "1 minute ago"),
    (timedelta(minutes=1, seconds=59), "1 minute ago"),
    (timedelta(minutes=2), "2 minutes ago"),
    (timedelta(minutes=59, seconds=59), "59 minutes ago"),
    (timedelta(minutes=60), "1 hour ago"),
    (timedelta(hours=1), "1 hour ago"),
    (timedelta(hours=1, minutes=59, seconds=59), "1 hour ago"),
    (timedelta(hours=2), "2 hours ago"),
    (timedelta(hours=23, minutes=59, seconds=59), "23 hours ago"),
    (timedelta(hours=24), "1 day ago"),
    (timedelta(days=1, hours=23, minutes=59, seconds=59), "1 day ago"),
    (timedelta(hours=48), "2 days ago"),
    (timedelta(days=29, hours=23, minutes=59, seconds=59), "29 days ago"),
    (timedelta(days=30), "1 month ago"),
    (timedelta(days=59, hours=23, minutes=59, seconds=59), "1 month ago"),
    (timedelta(days=60), "2 months ago"),
])
def test_humanize_time(inp, expected):
    # flexmock(datetime, now=datetime(year=2000, month=1, day=1))
    n = datetime.now()
    assert humanize_time(n - inp) == expected


def test_ordered_set():
    s = OrderedSet()
    s.append(1)
    assert s == [1]
    s.append(2)
    assert s == [1, 2]
    s.append(1)
    assert s == [2, 1]


# @pytest.mark.parametrize("inp,expected", [
#     ("aaa", ["aaa"]),
#     (
#         "qwe [01;34m asd [01;33mbnm",
#         ["qwe ", " asd ", "bnm"]
#     )
# ])
# def test_colorize_text(inp, expected):
#     got = colorize_text(inp)
#     for idx, c in enumerate(got):
#         if isinstance(c, str):
#             assert expected[idx] == c
#         else:
#             assert expected[idx] == c[1]


@pytest.mark.parametrize("inp,expected", [
    ("aaa", "aaa"),
    (
        "root:x:0:0:root:/root:/bin/b\x1b[01;31m\x1b[Ka\x1b[m\x1b[Ksh",
        "root:x:0:0:root:/root:/bin/bash"
    ),
    (
        "\x1b[0m\x1b[01;36mbin\x1b[0m\r\n\x1b[01;34mboot\x1b[0m\r\n\x1b[01;34mdev\x1b[0m\r\n",
        "bin\r\nboot\r\ndev\r\n"
    )
])
def test_strip_from_ansi_seqs(inp, expected):
    got = strip_from_ansi_esc_sequences(inp)
    assert got == expected



import pprint
import logging
import threading

import urwid
import urwidtrees
from urwid import BoxAdapter

from sen.exceptions import NotAvailableAnymore, NotifyError
from sen.tui.chunks.container import ContainerStatusWidget
from sen.tui.chunks.image import LayerWidget
from sen.tui.views.base import View
from sen.tui.widgets.graph import ContainerInfoGraph
from sen.tui.widgets.list.base import WidgetBase
from sen.tui.widgets.list.util import RowWidget, UnselectableRowWidget
from sen.tui.widgets.table import assemble_rows
from sen.tui.widgets.util import (
    SelectableText, get_map,
    ColorText, UnselectableListBox
)
from sen.util import log_traceback, humanize_bytes


logger = logging.getLogger(__name__)


class Process:
    """
    single process returned for container.stats() query

    so we can hash the object
    """
    def __init__(self, data):
        self.data = data

    @property
    def pid(self):
        return self.data["PID"]

    @property
    def ppid(self):
        return self.data["PPID"]

    @property
    def command(self):
        return self.data["COMMAND"]

    def __str__(self):
        return "[{}] {}".format(self.pid, self.command)

    def __repr__(self):
        return self.__str__()


class ProcessList:
    """
    util functions for process returned by container.stats()
    """

    def __init__(self, data):
        self.data = [Process(x) for x in data]
        self._nesting = {x.pid: [] for x in self.data}
        for x in self.data:
            try:
                self._nesting[x.ppid].append(x)
            except KeyError:
                pass

        logger.debug(pprint.pformat(self._nesting, indent=2))
        self._pids = [x.pid for x in self.data]
        self._pid_index = {x.pid: x for x in self.data}

    def get_parent_process(self, process):
        return self._pid_index.get(process.ppid, None)

    def get_root_process(self):
        # FIXME: error handling
        root_process = [x for x in self.data if x.ppid not in self._pids]
        return root_process[0]

    def get_first_child_process(self, process):
        try:
            return self._nesting[process.pid][0]
        except (KeyError, IndexError):
            return

    def get_last_child_process(self, process):
        try:
            return self._nesting[process.pid][-1]
        except (KeyError, IndexError):
            return

    def get_next_sibling(self, process):
        children = self._nesting.get(process.ppid, [])
        if len(children) <= 0:
            return None
        try:
            p = children[children.index(process) + 1]
        except IndexError:
            return
        return p

    def get_prev_sibling(self, process):
        children = self._nesting.get(process.ppid, [])
        if len(children) <= 0:
            return None
        logger.debug("prev of %s has children %s", process, children)
        prev_idx = children.index(process) - 1
        if prev_idx < 0:
            # when this code path is not present, tree navigation is seriously messed up
            return None
        else:
            return children[prev_idx]


class ProcessTreeBackend(urwidtrees.Tree):
    def __init__(self, data):
        """

        :param data: dict, response from container.top()
        """
        super().__init__()
        self.data = data
        self.process_list = ProcessList(data)
        self.root = self.process_list.get_root_process()

    def __getitem__(self, pos):
        logger.debug("do widget for %s", pos)
        return RowWidget([SelectableText(str(pos))])

    # Tree API
    def parent_position(self, pos):
        v = self.process_list.get_parent_process(pos)
        logger.debug("parent of %s is %s", pos, v)
        return v

    def first_child_position(self, pos):
        logger.debug("first child process for %s", pos)
        v = self.process_list.get_first_child_process(pos)
        logger.debug("first child of %s is %s", pos, v)
        return v

    def last_child_position(self, pos):
        v = self.process_list.get_last_child_process(pos)
        logger.debug("last child of %s is %s", pos, v)
        return v

    def next_sibling_position(self, pos):
        v = self.process_list.get_next_sibling(pos)
        logger.debug("next of %s is %s", pos, v)
        return v

    def prev_sibling_position(self, pos):
        v = self.process_list.get_prev_sibling(pos)
        logger.debug("prev of %s is %s", pos, v)
        return v


class ProcessTree(urwidtrees.TreeBox):
    def __init__(self, data):
        tree = ProcessTreeBackend(data)

        # We hide the usual arrow tip and use a customized collapse-icon.
        t = urwidtrees.ArrowTree(
            tree,
            arrow_att="tree",  # lines, tip
            icon_collapsed_att="tree",  # +
            icon_expanded_att="tree",  # -
            icon_frame_att="tree",  # [ ]
        )
        super().__init__(t)


class ContainerInfoView(WidgetBase, View):
    """
    display info about container
    """
    def __init__(self, ui, docker_container):
        self.walker = urwid.SimpleFocusListWalker([])
        super().__init__(ui, self.walker)

        self.docker_container = docker_container

        self.stop = threading.Event()

        self.view_widgets = []

    def refresh(self):
        self.view_widgets.clear()
        self.docker_container.refresh()

        self._basic_data()
        self._net()
        self._image()
        self._process_tree()
        self._resources()
        self._labels()
        self._logs()

        # we'll update listwalker in one step: changing it periodically can be race-y
        self.set_body(self.view_widgets)
        self.set_focus(0)

    @property
    def focused_docker_object(self):
        try:
            return self.focus.columns.widget_list[0].docker_image
        except AttributeError:
            return None

    def _basic_data(self):
        data = [
            [SelectableText("Id", maps=get_map("main_list_green")),
             SelectableText(self.docker_container.container_id)],
            [SelectableText("Status", maps=get_map("main_list_green")),
             ContainerStatusWidget(self.docker_container, nice_status=False)],
            [SelectableText("Created", maps=get_map("main_list_green")),
             SelectableText("{0}, {1}".format(self.docker_container.display_formal_time_created(),
                                              self.docker_container.display_time_created()))],
            [SelectableText("Command", maps=get_map("main_list_green")),
             SelectableText(self.docker_container.command)],
        ]
        # TODO: add exit code, created, started, finished, pid
        if self.docker_container.names:
            data.append(
                [SelectableText("Name", maps=get_map("main_list_green")),
                 SelectableText("".join(self.docker_container.names))],
            )

        if self.docker_container.size_root_fs:
            data.append(
                [SelectableText("Image Size", maps=get_map("main_list_green")),
                 SelectableText(humanize_bytes(self.docker_container.size_root_fs))])
        if self.docker_container.size_rw_fs:
            data.append(
                [SelectableText("Writable Layer Size", maps=get_map("main_list_green")),
                 SelectableText(humanize_bytes(self.docker_container.size_rw_fs))])

        self.view_widgets.extend(assemble_rows(data, ignore_columns=[1]))

    def _net(self):
        try:
            net = self.docker_container.net
        except NotAvailableAnymore:
            raise NotifyError("Container %s is not available anymore" % self.docker_container)
        ports = net.ports
        data = []
        if ports:
            data.extend([[SelectableText("")], [
                SelectableText("Host Port", maps=get_map("main_list_white")),
                SelectableText("Container Port", maps=get_map("main_list_white"))
            ]])
            for container_port, host_port in ports.items():
                if host_port and container_port:
                    data.append([
                        SelectableText(host_port), SelectableText(container_port)
                    ])

        ips = net.ips
        logger.debug(ips)
        if ips:
            data.extend([[SelectableText("")], [
                SelectableText("Network Name", maps=get_map("main_list_white")),
                SelectableText("IP Address", maps=get_map("main_list_white"))
            ]])
            for net_name, net_data in ips.items():
                a4 = net_data.get("ip_address4", "none")
                a6 = net_data.get("ip_address6", "")
                data.append([
                    SelectableText(net_name), SelectableText(a4)
                ])
                if a6:
                    data.append([
                        SelectableText(net_name), SelectableText(a6)
                    ])
        if data:
            self.view_widgets.extend(assemble_rows(data, dividechars=3, ignore_columns=[1]))

    def _image(self):
        if self.docker_container.image:
            self.view_widgets.append(RowWidget([SelectableText("")]))
            self.view_widgets.append(RowWidget([SelectableText("Image", maps=get_map("main_list_white"))]))
            self.view_widgets.append(RowWidget([LayerWidget(self.ui, self.docker_container.image)]))

    def _resources(self):
        self.view_widgets.append(RowWidget([SelectableText("")]))
        self.view_widgets.append(RowWidget([SelectableText("Resource Usage",
                                                     maps=get_map("main_list_white"))]))
        cpu_g = ContainerInfoGraph("graph_lines_cpu_tips", "graph_lines_cpu")
        mem_g = ContainerInfoGraph("graph_lines_mem_tips", "graph_lines_mem")
        blk_r_g = ContainerInfoGraph("graph_lines_blkio_r_tips", "graph_lines_blkio_r")
        blk_w_g = ContainerInfoGraph("graph_lines_blkio_w_tips", "graph_lines_blkio_w")
        net_r_g = ContainerInfoGraph("graph_lines_net_r_tips", "graph_lines_net_r")
        net_w_g = ContainerInfoGraph("graph_lines_net_w_tips", "graph_lines_net_w")

        cpu_label = ColorText("CPU ", "graph_lines_cpu_legend")
        cpu_value = ColorText("0.0 %", "graph_lines_cpu_legend")
        mem_label = ColorText("Memory ", "graph_lines_mem_legend")
        mem_value = ColorText("0.0 %", "graph_lines_mem_legend")
        blk_r_label = ColorText("I/O Read ", "graph_lines_blkio_r_legend")
        blk_r_value = ColorText("0 B", "graph_lines_blkio_r_legend")
        blk_w_label = ColorText("I/O Write ", "graph_lines_blkio_w_legend")
        blk_w_value = ColorText("0 B", "graph_lines_blkio_w_legend")
        net_r_label = ColorText("Net Rx ", "graph_lines_net_r_legend")
        net_r_value = ColorText("0 B", "graph_lines_net_r_legend")
        net_w_label = ColorText("Net Tx ", "graph_lines_net_w_legend")
        net_w_value = ColorText("0 B", "graph_lines_net_w_legend")
        self.view_widgets.append(urwid.Columns([
            BoxAdapter(cpu_g, 12),
            BoxAdapter(mem_g, 12),
            ("weight", 0.5, BoxAdapter(blk_r_g, 12)),
            ("weight", 0.5, BoxAdapter(blk_w_g, 12)),
            ("weight", 0.5, BoxAdapter(net_r_g, 12)),
            ("weight", 0.5, BoxAdapter(net_w_g, 12)),
            BoxAdapter(UnselectableListBox(urwid.SimpleFocusListWalker([
                UnselectableRowWidget([(12, cpu_label), cpu_value]),
                UnselectableRowWidget([(12, mem_label), mem_value]),
                UnselectableRowWidget([(12, blk_r_label), blk_r_value]),
                UnselectableRowWidget([(12, blk_w_label), blk_w_value]),
                UnselectableRowWidget([(12, net_r_label), net_r_value]),
                UnselectableRowWidget([(12, net_w_label), net_w_value]),
            ])), 12),
        ]))
        self.view_widgets.append(RowWidget([SelectableText("")]))

        @log_traceback
        def realtime_updates():
            g = self.docker_container.stats().response
            while True:
                try:
                    update = next(g)
                except Exception as ex:
                    if "Timeout" in ex.__class__.__name__:
                        logger.info("timeout when reading stats: %r", ex)
                        g = self.docker_container.stats().response
                        continue
                    logger.error("error while getting stats: %r", ex)
                    self.ui.notify_message("Error while getting stats: %s" % ex, level="error")
                    # TODO: if debug raise
                    break

                if self.stop.is_set():
                    break
                logger.debug(update)
                cpu_percent = update["cpu_percent"]
                cpu_value.text = "%.2f %%" % cpu_percent
                cpu_g.rotate_value(int(cpu_percent), max_val=100)

                mem_percent = update["mem_percent"]
                mem_current = humanize_bytes(update["mem_current"])
                mem_value.text = "%.2f %% (%s)" % (mem_percent, mem_current)
                mem_g.rotate_value(int(mem_percent), max_val=100)

                blk_read = update["blk_read"]
                blk_write = update["blk_write"]
                blk_r_value.text = humanize_bytes(blk_read)
                blk_w_value.text = humanize_bytes(blk_write)
                r_max_val = blk_r_g.rotate_value(blk_read, adaptive_max=True)
                w_max_val = blk_w_g.rotate_value(blk_write, adaptive_max=True)
                blk_r_g.set_max(max((r_max_val, w_max_val)))
                blk_w_g.set_max(max((r_max_val, w_max_val)))

                net_read = update["net_rx"]
                net_write = update["net_tx"]
                net_r_value.text = humanize_bytes(net_read)
                net_w_value.text = humanize_bytes(net_write)
                r_max_val = net_r_g.rotate_value(net_read, adaptive_max=True)
                w_max_val = net_w_g.rotate_value(net_write, adaptive_max=True)
                net_r_g.set_max(max((r_max_val, w_max_val)))
                net_w_g.set_max(max((r_max_val, w_max_val)))

        self.thread = threading.Thread(target=realtime_updates, daemon=True)
        self.thread.start()

    def _labels(self):
        if not self.docker_container.labels:
            return []
        data = []
        self.view_widgets.append(RowWidget([SelectableText("Labels", maps=get_map("main_list_white"))]))
        for label_key, label_value in self.docker_container.labels.items():
            data.append([SelectableText(label_key, maps=get_map("main_list_green")), SelectableText(label_value)])
        self.view_widgets.extend(assemble_rows(data, ignore_columns=[1]))

    def _process_tree(self):
        top = self.docker_container.top().response
        logger.debug(top)
        if top:
            self.view_widgets.append(RowWidget([SelectableText("")]))
            self.view_widgets.append(RowWidget([SelectableText("Process Tree",
                                                         maps=get_map("main_list_white"))]))
            self.view_widgets.append(BoxAdapter(ProcessTree(top), len(top)))

    def _logs(self):
        operation = self.docker_container.logs(follow=False, lines=10)
        if operation.response:
            l = []
            l.append(RowWidget([SelectableText("")]))
            l.append(RowWidget([SelectableText("Logs", maps=get_map("main_list_white"))]))
            l.extend([RowWidget([SelectableText(x)]) for x in operation.response.splitlines()])
            self.view_widgets.extend(l)

    def destroy(self):
        self.stop.set()



import math
import logging

import urwid


logger = logging.getLogger(__name__)


def find_max(list_of_lists):
    list_of_ints = [x[0] for x in list_of_lists]
    m = max(list_of_ints)
    try:
        return 2 ** int(math.log2(m) + 1)
    except ValueError:
        return 1


class ContainerInfoGraph(urwid.BarGraph):
    def __init__(self, fg, bg, graph_bg="graph_bg", bar_width=None):
        """
        create a very simple graph

        :param fg: attr for smoothing (fg needs to be set)
        :param bg: attr for bars (bg needs to be set)
        :param graph_bg: attr for graph background
        :param bar_width: int, width of bars
        """
        # satt smoothes graph lines
        satt = {(1, 0): fg}

        super().__init__(
            [graph_bg, bg],
            hatt=[fg],
            satt=satt,
        )
        if bar_width is not None:
            # breaks badly when set too high
            self.set_bar_width(bar_width)

    def render(self, size, focus=False):
        data, top, hlines = self._get_data(size)
        maxcol, maxrow = size
        if len(data) < maxcol:
            data += [[0] for x in range(maxcol - len(data))]
            self.set_data(data, top, hlines)
            logger.debug(data)
        return super().render(size, focus)

    def rotate_value(self, val, max_val=None, adaptive_max=False):
        """

        """
        data, _, _ = self.data
        data = data[1:] + [[int(val)]]
        if adaptive_max:
            max_val = find_max(data)

        self.set_data(data, max_val)
        return max_val

    def set_max(self, value):
        data, top, hlines = self.data
        self.set_data(data, value, hlines)




#!/usr/bin/env python3
"""
yes, this is python 3 ONLY project
"""

from __future__ import print_function

import sys

# let's be so gentle and print the error message even on python 2

if sys.version_info.major <= 2:
    error_message = """\
It looks like you are running sen with python 2. I'm sorry but sen is a python 3 only project.
Please see installation steps at official project page:

https://github.com/TomasTomecek/sen"""
    print(error_message, file=sys.stderr)
    sys.exit(2)

import argparse
import logging

import sen
from sen import set_logging
from sen.exceptions import TerminateApplication
from sen.tui.init import Application
from sen.util import get_log_file_path, log_last_traceback

logger = logging.getLogger("sen")


def main():
    parser = argparse.ArgumentParser(
        description="Terminal User Interface for Docker Engine"
    )
    parser.add_argument(
        "--yolo", "--skip-prompt-for-irreversible-action",
        action="store_true",
        default=False,
        help="Don't prompt when performing irreversible actions, a.k.a. YOLO!"
    )
    exclusive_group = parser.add_mutually_exclusive_group()
    exclusive_group.add_argument(
            "--debug", action="store_true", default=None,
            help="Set logging level to debug"
            )

    args = parser.parse_args()

    # !IMPORTANT! make sure that sen does NOT log via `logging.info` b/c it sets up root logger
    # and adds StreamHandler which causes to display logs on stdout which is definitely what we
    # don't want in a terminal app (thanks to Slavek Kabrda for explanation)
    if args.debug:
        set_logging(level=logging.DEBUG, path=get_log_file_path())
        logger.debug("sen loaded from %s", sen.__file__)
    else:
        set_logging(level=logging.INFO, path=get_log_file_path())

    logger.info("application started")

    try:
        app = Application(yolo=args.yolo)
    except TerminateApplication as ex:
        print("Error: {0}".format(str(ex)), file=sys.stderr)
        return 1

    try:
        app.run()
    except KeyboardInterrupt:
        print("Quitting on user request.")
        return 1
    except Exception as ex:  # pylint: disable=broad-except
        log_last_traceback()
        if args.debug:
            raise
        else:
            # TODO: improve this message to be more thorough
            print("There was an error during program execution, see logs for more info.")
            return 1
    return 0


if __name__ == "__main__":
    sys.exit(main())



"""
Container-specific chunks.
"""

import logging

from sen.tui.widgets.util import SelectableText, get_map


logger = logging.getLogger(__name__)


class ContainerStatusWidget(SelectableText):
    def __init__(self, docker_container, nice_status=True):
        markup, attr = get_container_status_markup(docker_container, nice_status=nice_status)
        super().__init__(markup, attr)


class ContainerOneLinerWidget(SelectableText):
    def __init__(self, ui, docker_container):
        self.ui = ui
        self.docker_container = docker_container
        super().__init__(get_basic_container_markup(docker_container))


def get_detailed_container_row(docker_container):
    row = []
    container_id = SelectableText(docker_container.short_id)
    row.append(container_id)

    commands = docker_container.command.split("\n")
    command_str = commands[0]
    if len(commands) > 0:
        command_str += "..."
    command = SelectableText(command_str, get_map(defult="main_list_ddg"))
    row.append(command)

    image = SelectableText(docker_container.image_name())
    row.append(image)

    row.append(ContainerStatusWidget(docker_container))

    name = SelectableText(docker_container.short_name)
    row.append(name)

    return row


def get_container_status_markup(docker_container, nice_status=True):
    if docker_container.running:
        attr_map = get_map("main_list_green")
    elif docker_container.status_created:
        attr_map = get_map("main_list_yellow")
    elif docker_container.exited_well:
        attr_map = get_map("main_list_orange")
    else:
        attr_map = get_map("main_list_red")
    if nice_status:
        return docker_container.nice_status, attr_map
    else:
        return docker_container.simple_status_cap, attr_map


def get_basic_container_markup(docker_container):
    text_markup = [docker_container.short_id, " "]

    markup, attr = get_container_status_markup(docker_container)
    text_markup.append((attr["normal"], markup))

    if docker_container.names:
        text_markup.append(" ")
        text_markup.append(("main_list_lg", docker_container.names[0]))

    return text_markup


"""
Definition of commands

This could be also split into two parts: generic framework part, application specific part
"""

import logging
import shlex


logger = logging.getLogger(__name__)

# command -> Class
commands_mapping = {}


class KeyNotMapped(Exception):
    pass


class NoSuchCommand(Exception):
    pass


class NoSuchOptionOrArgument(Exception):
    pass


# class decorator to register commands
def register_command(kls):
    commands_mapping[kls.name] = kls
    for a in kls.aliases:
        commands_mapping[a] = kls
    return kls


class CommandPriority:
    pass


class BackendPriority(CommandPriority):
    """ command takes long to execute """


class FrontendPriority(CommandPriority):
    """ command needs to be executed ASAP """


class SameThreadPriority(CommandPriority):
    """ run the task in same thread as UI """


def true_action(val=None):
    if val is not None:
        return val
    return True


class ArgumentBase:
    """
    Base class for arguments and options
    """
    def __init__(self, name, description, action=true_action, default=None):
        self.name = name
        self.description = description
        self.default = default
        self.action = action


class Option(ArgumentBase):
    """
    options alter behavior, are not positional
    """
    def __init__(self, name, description, action=true_action, aliases=None, default=None):
        super().__init__(name, description, action=action, default=default)
        self.aliases = aliases or []

    def __str__(self):
        return "{} default={} action=\"{}\" aliases={}".format(
            self.name, self.default, self.action, self.aliases
        )

    def __unicode__(self):
        return self.__str__()


class Argument(ArgumentBase):
    """
    arguments are positional
    """
    pass


def normalize_arg_name(name):
    return name.replace("-", "_")  # so we can access names-with-dashes


class ArgumentProcessor:
    """
    responsible for parsing given list of arguments
    """
    def __init__(self, options, arguments):
        """
        :param options: list of options
        :param arguments: list of arguments
        """
        self.given_arguments = {}
        self.options = {}
        for a in options:
            self.options[a.name] = a
            self.given_arguments[normalize_arg_name(a.name)] = a.default
            for alias in a.aliases:
                self.options[alias] = a
        for o in arguments:
            self.given_arguments[normalize_arg_name(o.name)] = o.default
        self.arguments = arguments
        logger.info("arguments = %s", arguments)
        logger.info("options = %s", options)

    def process(self, argument_list):
        """
        :param argument_list: list of str, input from user
        :return: dict:
            {"cleaned_arg_name": "value"}
        """
        arg_index = 0
        for a in argument_list:
            opt_and_val = a.split("=", 1)
            opt_name = opt_and_val[0]
            try:
                # option
                argument = self.options[opt_name]
            except KeyError:
                # argument
                try:
                    argument = self.arguments[arg_index]
                except IndexError:
                    logger.error("option/argument %r not specified", a)
                    raise NoSuchOptionOrArgument("No such option or argument: %r" % opt_name)
            logger.info("argument found: %s", argument)

            safe_arg_name = normalize_arg_name(argument.name)  # so we can access names-with-dashes

            logger.info("argument is available under name %r", safe_arg_name)

            if isinstance(argument, Argument):
                arg_index += 1
                value = (a, )
            else:
                try:
                    value = (opt_and_val[1], )
                except IndexError:
                    value = tuple()

            arg_val = argument.action(*value)

            logger.info("argument %r has value %r", safe_arg_name, arg_val)
            self.given_arguments[safe_arg_name] = arg_val
        return self.given_arguments


class CommandArgumentsGetter:
    def __init__(self, given_arguments):
        self.given_arguments = given_arguments

    def set_argument(self, arg_name, value):
        self.given_arguments[arg_name] = value

    def __getattr__(self, item):
        try:
            return self.given_arguments[item]
        except KeyError:
            # this is an error in code, not user error
            logger.error("no argument/option defined: %r", item)
            raise AttributeError("No such option or argument: %r" % item)


class Command:
    # command name, unique identifier, used also in prompt
    name = ""
    # message explaining what's about to happen
    pre_info_message = ""
    # message explaining what has happened
    post_info_message = ""
    # how long it takes to run the command - in which queue it should be executed
    priority = None
    # used in help message
    description = ""
    # define options
    options_definitions = []
    # define arguments
    arguments_definitions = []
    # command is available under these aliases
    aliases = []

    def __init__(self, ui=None, docker_backend=None, docker_object=None, buffer=None, size=None):
        """

        :param ui:
        :param docker_backend:
        :param docker_object:
        :param buffer:
        """
        logger.debug(
            "command %r initialized: ui=%r, docker_backend=%r, docker_object=%r, buffer=%r",
            self.name, ui, docker_backend, docker_object, buffer)
        self.ui = ui
        self.docker_backend = docker_backend
        self.docker_object = docker_object
        self.buffer = buffer
        self.size = size
        self.argument_processor = ArgumentProcessor(self.options_definitions,
                                                    self.arguments_definitions)
        self.arguments = None

    def process_args(self, arguments):
        """

        :param arguments: dict
        :return:
        """
        given_arguments = self.argument_processor.process(arguments)
        logger.info("given arguments = %s", given_arguments)
        self.arguments = CommandArgumentsGetter(given_arguments)

    def run(self):
        raise NotImplementedError()


class FrontendCommand(Command):
    priority = FrontendPriority()


class BackendCommand(Command):
    priority = BackendPriority()


class SameThreadCommand(Command):
    priority = SameThreadPriority()


class Commander:
    """
    Responsible for managing commands: it's up to workers to do the commands actually.
    """
    def __init__(self, ui, docker_backend):
        self.ui = ui
        self.docker_backend = docker_backend
        self.modifier_keys_pressed = []
        logger.debug("available commands: %s", commands_mapping)

    def get_command(self, command_input, docker_object=None, buffer=None, size=None):
        """
        return command instance which is the actual command to be executed

        :param command_input: str, command name and its args: "command arg arg2=val opt"
        :param docker_object:
        :param buffer:
        :param size: tuple, so we can call urwid.keypress(size, ...)
        :return: instance of Command
        """
        logger.debug("get command for command input %r", command_input)

        if not command_input:
            # noop, don't do anything
            return

        if command_input[0] in ["/"]:  # we could add here !, @, ...
            command_name = command_input[0]
            unparsed_command_args = shlex.split(command_input[1:])
        else:
            command_input_list = shlex.split(command_input)
            command_name = command_input_list[0]
            unparsed_command_args = command_input_list[1:]

        try:
            CommandClass = commands_mapping[command_name]
        except KeyError:
            logger.info("no such command: %r", command_name)
            raise NoSuchCommand("There is no such command: %s" % command_name)
        else:
            cmd = CommandClass(ui=self.ui, docker_backend=self.docker_backend,
                               docker_object=docker_object, buffer=buffer, size=size)
            cmd.process_args(unparsed_command_args)
            return cmd

    def get_command_input_by_key(self, key):
        logger.debug("get command input for key %r", key)

        modifier_keys = ["g"]  # FIXME: we should be able to figure this out from existing keybinds

        inp = "".join(self.modifier_keys_pressed) + key

        try:
            command_input = self.ui.current_buffer.get_keybinds()[inp]
        except KeyError:
            if key in modifier_keys:
                # TODO: inform user maybe
                self.modifier_keys_pressed.append(key)
                logger.info("modifier keys pressed: %s", self.modifier_keys_pressed)
                return
            else:
                logger.info("no such keybind: %r", inp)
                self.modifier_keys_pressed.clear()
                raise KeyNotMapped("No such keybind: %r." % inp)
        else:
            self.modifier_keys_pressed.clear()
            return command_input



"""
networking in docker backend
"""
from sen.util import graceful_chain_get


def extract_data_from_inspect(network_name, network_data):
    """
    :param network_name: str
    :param network_data: dict
    :return: dict:
        {
            "ip_address4": "12.34.56.78"
            "ip_address6": "ff:fa:..."
        }
    """
    a4 = None
    if network_name == "host":
        a4 = "127.0.0.1"
    n = {}
    a4 = graceful_chain_get(network_data, "IPAddress") or a4
    if a4:
        n["ip_address4"] = a4
    a6 = graceful_chain_get(network_data, "GlobalIPv6Address")
    if a6:
        n["ip_address4"] = a6
    return n


class NetData:
    def __init__(self, inspect_data):
        self.inspect_data = inspect_data
        self.net_settings = graceful_chain_get(self.inspect_data, "NetworkSettings")
        self._ports = None
        self._ips = None

    @property
    def ports(self):
        """
        :return: dict
            {
                # container -> host
                "1234": "2345"
            }
        """

        if self._ports is None:
            self._ports = {}
            if self.net_settings["Ports"]:
                for key, value in self.net_settings["Ports"].items():
                    cleaned_port = key.split("/")[0]
                    self._ports[cleaned_port] = graceful_chain_get(value, 0, "HostPort")
            # in case of --net=host, there's nothing in network settings, let's get it from "Config"
            exposed_ports_section = graceful_chain_get(self.inspect_data, "Config", "ExposedPorts")
            if exposed_ports_section:
                for key, value in exposed_ports_section.items():
                    cleaned_port = key.split("/")[0]
                    self._ports[cleaned_port] = None  # extremely docker specific
        return self._ports

    @property
    def ips(self):
        """
        :return: dict:
        {
            "default": {
                "ip_address4": "12.34.56.78"
                "ip_address6": "ff:fa:..."
            }
            "other": {
                ...
            }
        }
        """
        if self._ips is None:
            self._ips = {}
            default_net = extract_data_from_inspect("default", self.net_settings)
            if default_net:
                self._ips["default"] = default_net
            # this can be None
            networks = self.inspect_data["NetworkSettings"]["Networks"]
            if networks:
                for network_name, network_data in networks.items():
                    self._ips[network_name] = extract_data_from_inspect(network_name, network_data)

        return self._ips
